#! /usr/bin/env python

'''
To make the generation of WFSS simulated integrations easier,
combine the 4 stages of the simulator
(seed image generator, disperser, dark prep, obervation generator)
into a single script.

Inputs:
paramfiles - List of yaml filenames. These files should be
             inputs to the simulator. For details on the
             information contained in the yaml files, see
             the readme file associated with the mirage
             github repo:
             https://github.com/spacetelescope/mirage.git

crossing_filter - Name of the crossing filter to be used in
                  conjunction with the grism. All longwave
                  channel filter names are valid entries

module - Name of the NIRCam module to use for the simulation.
         Can be 'A' or 'B'

direction - Dispersion direction. Can be along rows ('R') or
            along columns ('C')

override_dark - If you wish to use a dark current file that
                has already gone through the dark_prep step
                of the pipeline and wish to use that for the
                simulation, set override_dark equal to the
                dark's filename. The dark_prep step will then
                be skipped.

HISTORY:
15 November 2017 - created, Bryan Hilbert
13 July 2018 - updated for name change to Mirage, Bryan Hilbert
'''

import copy
import os
import sys
import argparse
import yaml

import numpy as np
from astropy.io import fits
from NIRCAM_Gsim.grism_seed_disperser import Grism_seed
import pysiaf
from scipy.stats import sigmaclip

from .catalogs import spectra_from_catalog
from .seed_image import catalog_seed_image
from .dark import dark_prep
from .ramp_generator import obs_generator
from .utils import backgrounds, read_fits
from .utils.flux_cal import fluxcal_info
from .utils.constants import CATALOG_YAML_ENTRIES, MEAN_GAIN_VALUES, NIRISS_GRISM_THROUGHPUT_FACTOR
from .utils.utils import expand_environment_variable, get_filter_throughput_file

from .yaml import yaml_update

NIRCAM_GRISM_CROSSING_FILTERS = ['F322W2', 'F277W', 'F356W', 'F444W', 'F250M', 'F300M',
                                 'F335M', 'F360M', 'F410M', 'F430M', 'F323N', 'F405N',
                                 'F466N', 'F470N']
NIRISS_GRISM_CROSSING_FILTERS = ['F200W', 'F150W', 'F140M', 'F158M', 'F115W', 'F090W']


class WFSSSim():
    def __init__(self, paramfiles, SED_file=None, SED_normalizing_catalog_column=None,
                 final_SED_file=None, SED_dict=None, save_dispersed_seed=True, source_stamps_file=None,
                 extrapolate_SED=True, override_dark=None, disp_seed_filename=None, offline=False,
                 create_continuum_seds=True):

        # Set the MIRAGE_DATA environment variable if it is not
        # set already. This is for users at STScI.
        self.env_var = 'MIRAGE_DATA'
        self.datadir = os.environ.get(self.env_var)
        if self.datadir is None:
            raise ValueError(("WARNING: {} environment variable is not set."
                              "This must be set to the base directory"
                              "containing the darks, cosmic ray, PSF, etc"
                              "input files needed for the simulation."
                              "These files must be downloaded separately"
                              "from the Mirage package.".format(self.env_var)))

        # Make sure the input param file(s) is a list
        self.paramfiles = paramfiles
        if isinstance(paramfiles, str):
            self.paramfiles = [self.paramfiles]

        # Set the user-input parameters
        self.create_continuum_seds = create_continuum_seds
        self.SED_file = SED_file
        self.SED_dict = SED_dict
        self.SED_normalizing_catalog_column = SED_normalizing_catalog_column
        self.final_SED_file = final_SED_file
        self.override_dark = override_dark
        self.save_dispersed_seed = save_dispersed_seed
        self.source_stamps_file = source_stamps_file
        self.disp_seed_filename = disp_seed_filename
        self.extrapolate_SED = extrapolate_SED
        self.fullframe_apertures = ["NRCA5_FULL", "NRCB5_FULL", "NIS_CEN"]
        self.offline = offline

        # Make sure the right combination of parameter files and SED file
        # are given
        self.param_checks()

        # Attempt to find the crossing filter and dispersion direction
        # from the input paramfiles. Adjust any imaging mode parameter
        # files to have the mode set to wfss. This will ensure the seed
        # images will be the proper (expanded) dimensions
        self.paramfiles = self.find_param_info()

        # Make sure inputs are correct
        self.check_inputs()

        # Determine whether Mirage will create a final hdf5 file that
        # contains all sources. If only one paramfile is provided, this
        # must be done.
        if len(self.paramfiles) == 1:
            self.create_continuum_seds = True
        else:
            # If spectra are provided, this step must be done
            if self.SED_file is not None or self.SED_dict is not None:
                self.create_continuum_seds = True
            else:
                # If multiple paramfiles are given and no input spectra,
                # then we leave the choice up to the user, as either case
                # is ok.
                pass

    def create(self):
        """MAIN FUNCTION"""

        # Loop over the yaml files and create
        # a direct seed image for each
        imseeds = []
        ptsrc_seeds = []
        galaxy_seeds = []
        extended_seeds = []
        for pfile in self.paramfiles:
            print('Running catalog_seed_image for {}'.format(pfile))
            cat = catalog_seed_image.Catalog_seed(offline=self.offline)
            cat.paramfile = pfile
            cat.make_seed()
            imseeds.append(cat.seed_file)
            ptsrc_seeds.append(cat.ptsrc_seed_filename)
            galaxy_seeds.append(cat.galaxy_seed_filename)
            extended_seeds.append(cat.extended_seed_filename)

            # Save the flat field file associated with the wfss yaml so
            # that we can apply it to the dispersed seed image if requested
            if pfile == self.wfss_yaml:
                self.flatfield = cat.flatfield

            # If Mirage is going to produce an hdf5 file of spectra,
            # then we only need a single direct seed image. Note that
            # find_param_info() has reordered the list such that the
            # wfss mode yaml file will be examined first.
            if self.create_continuum_seds:
                break

        # Create hdf5 file with spectra of all sources if requested.
        if self.create_continuum_seds:
            det_name = cat.params['Readout']['array_name'].split('_')[0]
            self.SED_file = spectra_from_catalog.make_all_spectra(self.catalog_files, input_spectra=self.SED_dict,
                                                                  input_spectra_file=self.SED_file,
                                                                  extrapolate_SED=self.extrapolate_SED,
                                                                  output_filename=self.final_SED_file,
                                                                  normalizing_mag_column=self.SED_normalizing_catalog_column,
                                                                  module=self.module, detector=det_name)

        # Location of the configuration files needed for dispersion
        loc = os.path.join(self.datadir, "{}/GRISM_{}/".format(self.instrument,
                                                               self.instrument.upper()))

        # Determine the name of the background file to use, as well as the
        # orders to disperse.
        if self.instrument == 'nircam':
            dmode = 'mod{}_{}'.format(self.module, self.dispersion_direction)
            if self.params['simSignals']['use_dateobs_for_background']:
                print("Generating background spectrum for observation date: {}".format(self.params['Output']['date_obs']))
                back_wave, back_sig = backgrounds.day_of_year_background_spectrum(self.params['Telescope']['ra'],
                                                                                  self.params['Telescope']['dec'],
                                                                                  self.params['Output']['date_obs'])
            else:
                if isinstance(self.params['simSignals']['bkgdrate'], str):
                    if self.params['simSignals']['bkgdrate'].lower() in ['low', 'medium', 'high']:
                        print("Generating background spectrum based on requested level of: {}".format(self.params['simSignals']['bkgdrate']))
                        back_wave, back_sig = backgrounds.low_med_high_background_spectrum(self.params, self.detector,
                                                                                           self.module)
                    else:
                        raise ValueError("ERROR: Unrecognized background rate. Must be one of 'low', 'medium', 'high'")
                else:
                    raise ValueError(("ERROR: WFSS background rates must be one of 'low', 'medium', 'high', "
                                      "or use_dateobs_for_background must be True "))

        elif self.instrument == 'niriss':
            dmode = 'GR150{}'.format(self.dispersion_direction)
            background_file = "{}_{}_medium_background.fits".format(self.crossing_filter.lower(),
                                                                    dmode.lower())

            if isinstance(self.params['simSignals']['bkgdrate'], str):
                if self.params['simSignals']['bkgdrate'].lower() in ['low', 'medium', 'high']:
                    #scaling_factor = backgrounds.niriss_background_scaling(self.params, self.detector, self.module)

                    usefilt = 'pupil'

                    siaf_instance = pysiaf.Siaf('niriss')[self.params['Readout']['array_name']]
                    vegazp, photflam, photfnu, pivot_wavelength = fluxcal_info(self.params, usefilt, self.detector, self.module)

                    if os.path.split(self.params['Reffiles']['filter_throughput'])[1] == 'placeholder.txt' or self.params['Reffiles']['filter_throughput'] == 'config':
                        filter_file = get_filter_throughput_file(self.instrument, self.params['Readout'][usefilt])
                    else:
                        filter_file = self.params['Reffiles']['filter_throughput']

                    scaling_factor = backgrounds.calculate_background(self.params['Telescope']['ra'],
                                                                      self.params['Telescope']['dec'],
                                                                      filter_file,
                                                                      self.params['simSignals']['use_dateobs_for_background'],
                                                                      MEAN_GAIN_VALUES['niriss'], siaf_instance,
                                                                      level=self.params['simSignals']['bkgdrate'])

                    # Having the grism in the beam reduces the throughput by 20%.
                    # Mulitply that into the scaling factor
                    scaling_factor *= NIRISS_GRISM_THROUGHPUT_FACTOR

                    # Translate from ADU/sec/pix to e-/sec/pix since that is
                    # what the disperser works with
                    scaling_factor *= MEAN_GAIN_VALUES['niriss']

                else:
                    raise ValueError("ERROR: Unrecognized background rate. String value must be one of 'low', 'medium', 'high'")
            elif np.isreal(self.params['simSignals']['bkgdrate']):
                # The bkgdrate entry in the input yaml file is described as
                # the desired signal in ADU/sec/pixel IN A DIRECT IMAGE
                # Since we want e-/sec/pixel here for the disperser, multiply
                # by the gain as well as the throughput factor for the grism.
                scaling_factor = self.params['simSignals']['bkgdrate'] * MEAN_GAIN_VALUES['niriss'] * NIRISS_GRISM_THROUGHPUT_FACTOR

        # Default to extracting all orders
        orders = None

        # Call the disperser separately for each type of object: point sources
        # galaxies, extended objects
        disp_seed = np.zeros((cat.ffsize, cat.ffsize))
        background_done = False
        for seed_files in [ptsrc_seeds, galaxy_seeds, extended_seeds]:
            if seed_files[0] is not None:
                dispersed_objtype_seed = Grism_seed(seed_files, self.crossing_filter,
                                                    dmode, config_path=loc, instrument=self.instrument.upper(),
                                                    extrapolate_SED=self.extrapolate_SED, SED_file=self.SED_file,
                                                    SBE_save=self.source_stamps_file)
                dispersed_objtype_seed.observation(orders=orders)
                dispersed_objtype_seed.disperse(orders=orders)
                # Only include the background in one of the object type seed images
                if not background_done:
                    if self.instrument == 'nircam':
                        background_image = dispersed_objtype_seed.disperse_background_1D([back_wave, back_sig])
                        dispersed_objtype_seed.finalize(Back=background_image, BackLevel=None)
                    else:
                        # BackLevel is used as such: background / max(background) * BackLevel
                        # So we need to either set BackLevel equal to the requested level
                        # NOT THE RATIO OF THAT TO MEDIUM, or we need to open the background
                        # file and multiply it by the ratio of the requested level to medium.
                        # The former isn't quite correct because it'll be scaling the maximum
                        # value in the image to "low" or "high", rather than the median
                        full_background_file = os.path.join(loc, background_file)
                        background_image = fits.getdata(full_background_file)

                        # Before scaling the background image by the scaling_factor
                        # we need to normalize by the sigma-clipped mean value. This is
                        # because the background files were produced and scaled to the
                        # ETC "medium" level at some arbirtrary pointing, but the
                        # "medium" level is pointing-dependent. Current background files
                        # are scaled such that the "medium" value from the ETC is the
                        # sigma-clipped mean value.
                        clip, lo, hi = sigmaclip(background_image, low=3, high=3)
                        background_mean = np.mean(clip)
                        background_image = background_image / background_mean * scaling_factor
                        dispersed_objtype_seed.finalize(Back=background_image, BackLevel=None)

                    background_done = True
                else:
                    dispersed_objtype_seed.finalize()
                disp_seed += dispersed_objtype_seed.final

        # Disperser output is always full frame. Remove the signal from
        # the refrence pixels now since we know exactly where they are
        disp_seed[0:4, :] = 0.
        disp_seed[2044:, :] = 0.
        disp_seed[:, 0:4] = 0.
        disp_seed[:, 2044:] = 0.

        # Crop to the requested subarray if necessary
        if cat.params['Readout']['array_name'] not in self.fullframe_apertures:
            print("Subarray bounds: {}".format(cat.subarray_bounds))
            print("Dispersed seed image size: {}".format(disp_seed.shape))
            disp_seed = self.crop_to_subarray(disp_seed, cat.subarray_bounds)

            # Segmentation map will be centered in a frame that is larger
            # than full frame by a factor of sqrt(2), so crop appropriately
            print("Need to make this work for subarrays...")
            segy, segx = cat.seed_segmap.shape
            dx = int((segx - 2048) / 2)
            dy = int((segy - 2048) / 2)
            segbounds = [cat.subarray_bounds[0] + dx, cat.subarray_bounds[1] + dy,
                         cat.subarray_bounds[2] + dx, cat.subarray_bounds[3] + dy]
            cat.seed_segmap = self.crop_to_subarray(cat.seed_segmap, segbounds)

        # If it is a NIRCam observation, multiply the dispsersed seed
        # image by the flat field. For NIRISS, the flat was multiplied
        # in to the direct seed image, as they need to have their
        # occulting spots in the direct seed image.
        if self.instrument == 'nircam':
            disp_seed *= self.flatfield

        # Save the dispersed seed image if requested
        # Save in units of e/s, under the idea that this should be a
        # "perfect" noiseless view of the scene that does not depend on
        # detector effects, such as gain.
        if self.save_dispersed_seed:
            self.save_dispersed_seed_image(disp_seed)

        # Convert seed image to ADU/sec to be consistent
        # with other simulator outputs
        if self.instrument == 'niriss':
            gain = MEAN_GAIN_VALUES['niriss']
        elif self.instrument == 'nircam':
            gain = MEAN_GAIN_VALUES['nircam']['lw{}'.format(self.module.lower())]

        disp_seed /= gain

        # Update seed image header to reflect the
        # division by the gain
        cat.seedinfo['units'] = 'ADU/sec'

        # Prepare dark current exposure if
        # needed.
        if self.override_dark is None:
            d = dark_prep.DarkPrep(offline=self.offline)
            d.paramfile = self.wfss_yaml
            d.prepare()
            obslindark = d.prepDark
        else:
            self.read_dark_product()
            obslindark = self.darkPrep

        # Using the first of the imaging seed image yaml
        # files as a base, adjust to create the yaml file
        # for the creation of the final dispersed
        # integration

        # I think we won't need this anymore assuming that
        # one of the input yaml files is for wfss mode
        #y = yaml_update.YamlUpdate()
        #y.file = self.paramfiles[0]
        #if self.instrument == 'nircam':
        #    y.filter = self.crossing_filter
        #    y.pupil = 'GRISM' + self.direction
        #elif self.instrument == 'niriss':
        #    y.filter = 'GR150' + self.direction
        #    y.pupil = self.crossing_filter
        #y.outname = ("wfss_dispersed_{}_{}.yaml"
        #             .format(dmode, self.crossing_filter))
        #y.run()

        # Combine into final observation
        obs = obs_generator.Observation(offline=self.offline)
        obs.linDark = obslindark
        obs.seed = disp_seed
        obs.segmap = cat.seed_segmap
        obs.seedheader = cat.seedinfo
        #obs.paramfile = y.outname
        obs.paramfile = self.wfss_yaml
        obs.create()

    def param_checks(self):
        """Check parameter file inputs"""
        #if ((len(self.paramfiles) < 2) and (self.SED_file is None)):
        #    raise ValueError(("WARNING: Only one parameter file provided and no SED file. More "
        #                      "yaml files or an SED file needed in order to disperse."))

        if ((len(self.paramfiles) > 1) and (self.SED_file is not None)):
            raise ValueError(("WARNING: When using an SED file, you must provide only one parameter file."))

    def read_dark_product(self):
        # Read in dark product that was produced
        # by dark_prep.py
        self.prepDark = read_fits.Read_fits()
        self.prepDark.file = self.override_dark
        self.prepDark.read_astropy()

    def check_inputs(self):
        """Make sure input parameters are acceptible"""

        # ###################Instrument Name##################
        if self.instrument not in ['nircam', 'niriss']:
            self.invalid('instrument', self.instrument)

        # ###################Module Name##################
        if self.instrument == 'nircam':
            if self.module not in ['A', 'B']:
                self.invalid('module', self.module)
            else:
                self.module = self.module.upper()

            # ###################Crossing Filter##################
            if self.crossing_filter not in NIRCAM_GRISM_CROSSING_FILTERS:
                self.invalid('crossing_filter', self.crossing_filter)

        elif self.instrument == 'niriss':
            if self.crossing_filter not in NIRISS_GRISM_CROSSING_FILTERS:
                self.invalid('crossing_filter', self.crossing_filter)

        # ###################Dispersion Direction##################
        if self.dispersion_direction not in ['R', 'C']:
            self.invalid('dispersion_direction', self.dispersion_direction)

        # ###################Dark File to Use##################
        if self.override_dark is not None:
            avail = os.path.isfile(self.override_dark)
            if not avail:
                raise FileNotFoundError(("WARNING: {} does not exist."
                                         .format(self.override_dark)))

    def find_param_info(self):
        """Extract dispersion direction and crossing filter from the input
        param files"""
        yamls_to_disperse = []
        self.catalog_files = []
        wfss_files_found = 0
        for i, pfile in enumerate(self.paramfiles):
            with open(pfile, 'r') as infile:
                params = yaml.safe_load(infile)

            cats = [params['simSignals'][cattype] for cattype in CATALOG_YAML_ENTRIES if 'tso_' not in cattype]
            cats = [e for e in cats if e.lower() != 'none']
            self.catalog_files.extend(cats)

            if i == 0:
                self.instrument = params['Inst']['instrument'].lower()
                if self.instrument == 'niriss':
                    self.module = 'N'
                    self.detector = 'NIS'
                elif self.instrument == 'nircam':
                    self.module = params['Readout']['array_name'][3]
                    self.detector = params['Readout']['array_name'][0:5]

            if params['Inst']['mode'].lower() == 'wfss':
                self.wfss_yaml = copy.deepcopy(pfile)
                self.params = params

                # Only 1 input yaml file should be for wfss mode
                wfss_files_found += 1
                if wfss_files_found == 2:
                    raise ValueError("WARNING: only one of the parameter files can be WFSS mode.")
                filter_name = params['Readout']['filter']
                pupil_name = params['Readout']['pupil']
                dispname = ('{}_dispersed_seed_image.fits'.format(params['Output']['file'].split('.fits')[0]))
                self.default_dispersed_filename = os.path.join(params['Output']['directory'], dispname)

                # In reality, the grism elements are in NIRCam's pupil wheel, and NIRISS's
                # filter wheel. But in the APT xml file, NIRISS grisms are in the pupil
                # wheel and the crossing filter is listed in the filter wheel. At that
                # point, NIRISS and NIRCam are consistent, so let's keep with this reversed
                # information
                if self.instrument == 'niriss':
                    self.crossing_filter = pupil_name.upper()
                    self.dispersion_direction = filter_name[-1].upper()
                elif self.instrument == 'nircam':
                    self.crossing_filter = filter_name.upper()
                    self.dispersion_direction = pupil_name[-1].upper()
                # Prepend the wfss yaml file to the list, so that it is first
                yamls_to_disperse.insert(0, pfile)

            elif params['Inst']['mode'].lower() in ['imaging', 'pom']:
                # If the other yaml files are for imaging mode, we need to update them to
                # be wfss mode so that the resulting seed images have the correct dimensions.
                # Save these modified yaml files to new files.
                params['Inst']['mode'] = 'wfss'
                params['Output']['grism_source_image'] = True
                outdir, basename = os.path.split(pfile)
                modified_file = os.path.join(outdir, 'tmp_update_to_wfss_mode_{}'.format(basename))
                with open(modified_file, 'w') as output:
                    yaml.dump(params, output, default_flow_style=False)
                yamls_to_disperse.append(modified_file)

        if wfss_files_found == 0:
            raise ValueError(("WARNING: No WFSS mode parameter files found. One of the parameter "
                              "files must be wfss mode in order to define grism and crossing filter."))
        return yamls_to_disperse

    """
    def generate_background_spectrum(self, ra, dec, day_of_year=None):
        Call jwst_backgrounds in order to produce an estimate of background
        versus wavelength for a particular pointing

        background = jbt.background(ra, dec, 4.)
        # If the user wants a background signal from a particular day,
        # then extract that array here
        if self.params['simSignals']['use_dateobs_for_background']:
            obsdate = datetime.datetime.strptime(self.params['Output']['date_obs'], '%Y-%m-%d')
            obs_dayofyear = obsdate.timetuple().tm_yday
            if obs_dayofyear not in background.bkg_data['calendar']:
                raise ValueError(("ERROR: The requested RA, Dec is not observable on {}. Either "
                                  "specify a different day, or set simSignals:use_dateobs_for_background "
                                  "to False.".format(self.params['Output']['date_obs'])))
            #match = obs_dayofyear == background.bkg_data['calendar']
            #back_wave = background.bkg_data['wave_array']
            #back_sig = background.bkg_data['total_bg'][match, :]
            background_file = 'background.txt'
            background = jbt.get_background(ra, dec, 4., plot_background=False, plot_bathtub=False,
                                            thisday=day_of_year, write_background=True, write_bathtub=False,
                                            background_file=background_file)
        else:
            # If the user wants a background chosen from the three general
            # levels (low, medium, of high) then we need to examine the
            # returned background spectra for all days, determine from the
            # histogram (of background values at the pivot wavelength) what
            # "low", "medium", and "high" correspond to, and then pick out
            # the spectrum from a day that is close to the requested level
            vegazp, photflam, photfnu, pivot_wavelength = fluxcal_info(self.params, usefilt, detector, module)
            wave_diff = np.abs(background.bkg_data['wave_array'] - pivot_wavelength)
            bkgd_wave = np.where(wave_diff == np.min(wave_diff))[0][0]
            bkgd_at_pivot = background.bkg_data['total_bg'][:, bkgd_wave]

            # Now sort and determine the low/medium/high levels
            low, medium, high = find_low_med_high(bkgd_at_pivot)

            # Find the value based on the level in the yaml file
            background_level = self.params['simSignals']['bkgdrate'].lower()
            if background_level == "low":
                level_value = low
            elif background_level == "medium":
                level_value = medium
            elif background_level == "high":
                level_value = high
            else:
                raise ValueError(("ERROR: Unrecognized background value: {}. Must be low, mediumn, or high"
                                  .format(self.params['simSignals']['bkgdrate'])))

            # Find the day with the background at the pivot wavelength that
            # is closest to the value associated with the requested level
            diff = np.abs(bkgd_at_pivot - level_value)
            mindiff = np.where(diff == np.min(diff))[0][0]
            background_spec = background.bkg_data['total_bg'][mindiff, :]

        return background_file
    """

    def read_param_file(self, file):
        """
        Read in yaml simulator parameter file

        Parameters:
        -----------
        file -- Name of a yaml file in the proper format
                for mirage

        Returns:
        --------
        Nested dictionary with the yaml file's contents
        """
        import yaml
        try:
            with open(file, 'r') as infile:
                data = yaml.load(infile)
        except (FileNotFoundError, IOError) as e:
            print(e)

    def read_gain_file(self, file):
        """
        Read in CRDS-formatted gain reference file

        Paramters:
        ----------
        file -- Name of gain reference file

        Returns:
        --------
        Detector gain map (2d numpy array)
        """
        try:
            with fits.open(file) as h:
                image = h[1].data
                header = h[0].header
        except (FileNotFoundError, IOError) as e:
            print(e)

        mngain = np.nanmedian(image)

        # Set pixels with a gain value of 0 equal to mean
        image[image == 0] = mngain
        # Set any pixels with non-finite values equal to mean
        image[~np.isfinite(image)] = mngain
        return image, header

    def crop_to_subarray(self, data, bounds):
        """
        Crop the given full frame array down to the appropriate
        subarray size and location based on the requested subarray
        name.

        Parameters:
        -----------
        data -- 2d numpy array. Full frame image. (2048 x 2048)
        bounds -- 4-element list containing the full frame indices that
                  define the position of the subarray.
                  [xstart, ystart, xend, yend]

        Returns:
        --------
        Cropped 2d numpy array
        """
        yl, xl = data.shape
        valid = [False, False, False, False]
        valid = [(b >= 0 and b < xl) for b in bounds[0:3:2]]
        validy = [(b >= 0 and b < yl) for b in bounds[1:4:2]]
        valid.extend(validy)

        if all(valid):
            return data[bounds[1]:bounds[3] + 1, bounds[0]:bounds[2] + 1]
        else:
            raise ValueError(("WARNING: subarray bounds are outside the "
                              "dimensions of the input array."))

    def read_subarr_defs(self, subfile):
        # read in the file that contains a list of subarray
        # names and positions on the detector
        try:
            subdict = ascii.read(subfile, data_start=1, header_start=0)
            return subdict
        except (FileNotFoundError, IOError) as e:
            print(e)

    def get_subarr_bounds(self, subname, sdict):
        # find the bounds of the requested subarray
        if subname in sdict['AperName']:
            mtch = subname == sdict['AperName']
            bounds = [sdict['xstart'].data[mtch][0], sdict['ystart'].data[mtch][0],
                      sdict['xend'].data[mtch][0], sdict['yend'].data[mtch][0]]
            return bounds
        else:
            raise ValueError(("WARNING: {} is not a subarray aperture name present "
                              "in the subarray definition file.".format(subname)))

    def invalid(self, field, value):
        raise ValueError(("WARNING: invalid value for {}: {}"
                          .format(field, value)))

    def save_dispersed_seed_image(self, seed_image):
        """Save the dispersed seed image"""
        primary_hdu = fits.PrimaryHDU()
        image_hdu = fits.ImageHDU(seed_image)
        hdu_list = fits.HDUList([primary_hdu, image_hdu])
        hdu_list[0].header['units'] = 'e/sec'
        hdu_list[1].header['units'] = 'e/sec'
        if self.disp_seed_filename is None:
            self.disp_seed_filename = self.default_dispersed_filename
        hdu_list.writeto(self.disp_seed_filename, overwrite=True)
        print(("Dispersed seed image saved to {}".format(self.disp_seed_filename)))

    def add_options(self, parser=None, usage=None):
        if parser is None:
            parser = argparse.ArgumentParser(usage=usage, description=("Wrapper for the creation of"
                                                                       " WFSS simulated exposures."))
        parser.add_argument("paramfiles", help=('List of files describing the input parameters and '
                                                'instrument settings to use. (YAML format).'), nargs='+')
        parser.add_argument("--crossing_filter", help=("Name of crossing filter to use in conjunction "
                                                       "with the grism."), default=None)
        parser.add_argument("--module", help="NIRCam module to use for simulation. Use 'A' or 'B'",
                            default=None)
        parser.add_argument("--direction", help=("Direction of dispersion (along rows or along columns). "
                                                 "Use 'R' or 'C'"), default=None)
        parser.add_argument("--override_dark", help=("If supplied, skip the dark preparation step and "
                                                     "use the supplied dark to make the exposure"),
                            default=None)
        parser.add_argument("--extrapolate_SED", help=("If true, the SED created from the filter-averaged "
                                                       "magnitudes will be extrapolated to fill the "
                                                       "wavelngth range of the grism"), action='store_true')
        return parser


if __name__ == '__main__':

    usagestring = ('USAGE: wfss_simualtor.py file1.yaml file2.yaml --crossing_filter F444W --direction R '
                   '--module A')

    obs = WFSSSim()
    parser = obs.add_options(usage=usagestring)
    args = parser.parse_args(namespace=obs)
    obs.create()



#! /usr/bin/env python

"""Tests for ``catalog_seed_generator.py`` module

Authors
-------

    Bryan Hilbert

Use
---

    These tests can be run via the command line:

    ::

        pytest -s test_catalog_seed_generator.py
"""
from astropy.table import Table
import numpy as np
import os
import webbpsf

from mirage.seed_image import catalog_seed_image

# Determine if tests are being run on Travis
ON_TRAVIS = 'travis' in os.path.expanduser('~')
if ON_TRAVIS:
    env_path = os.path.expandvars('$CONDA_PREFIX')
    os.environ["WEBBPSF_PATH"] = os.path.join(env_path, 'share/webbpsf-data')


def create_dummy_psf_grid():
    """Use webbpsf to create a griddedPSFModel object"""
    nrc = webbpsf.NIRCam()
    nrc.filter = 'F200W'
    nrc.detector = 'NRCB1'
    grid = nrc.psf_grid(num_psfs=1, all_detectors=False, oversample=1, fov_pixels=301,
                        add_distortion=False, save=False)
    return grid


def test_overlap_coordinates_full_frame():
    """Test the function that calculates the coordinates for the
    overlap between two stamp images
    """
    # Create instance of Catalog_seed
    seed = catalog_seed_image.Catalog_seed(offline=True)

    # Create a point sources table
    tab = Table()
    tab['index'] = np.arange(10).astype(np.int) + 1
    tab['pixelx'] = [1000, 50, 1000, 1900, -0.5, 1033.587, 622.63, 1523.75, 1013.413, 1124.371]
    tab['pixely'] = [1000, 400, 25, 2000, -0.5, 1023.425, 1024.223, 1013.25, 223.575, 1822.72]
    tab['RA_degrees'] = [12.008729, 11.999914, 12.003422, 11.995729, 12.000036, 11.99918, 12.008729,
                         11.999914, 12.003422, 11.995729]
    tab['Dec_degrees'] = [-0.00885006, 1.7231344e-09, -1.5512744e-05, -4.9949034e-05, -0.006866415,
                          0.0068373946, -0.00885006, 1.7231344e-09, -1.5512744e-05, -4.9949034e-05]
    tab['magnitude'] = np.repeat(14, 10)
    tab['countrate_e/s'] = np.repeat(161630.72, 10)
    tab['counts_per_frame_e'] = np.repeat(1735391.9, 10)

    expected_results = []
    expected_results.append((850, 1151, 850, 1151, 0, 301, 0, 301))
    expected_results.append((0, 201, 250, 551, 100, 301, 0, 301))
    expected_results.append((850, 1151, 0, 176, 0, 301, 125, 301))
    expected_results.append((1750, 2048, 1850, 2048, 0, 298, 0, 198))

    stamp_dims = (301, 301)
    stamp_x = 150
    stamp_y = 150

    # Check the basic function
    aperture_dims = (2048, 2048)
    for index in range(4):
        results = seed.cropped_coords(tab[index]['pixelx'], tab[index]['pixely'], aperture_dims, stamp_x,
                                      stamp_y, stamp_dims, ignore_detector=False)
    assert results == expected_results[index]

    # Check the function that wraps around the most basic function
    seed.ffsize = 2048
    seed.subarray_bounds = [0, 0, 2047, 2047]
    for index in range(4):
        results = seed.create_psf_stamp_coords(tab[index]['pixelx'], tab[index]['pixely'], stamp_dims,
                                               stamp_x, stamp_y, coord_sys='full_frame',
                                               ignore_detector=False)
        ijkl = []
        for ele in results[4:]:
            ijkl.extend(list(ele))
        ijkl = tuple(ijkl)
        assert ijkl == expected_results[index]

    # Create dummy PSF library
    seed.psf_library = create_dummy_psf_grid()
    seed.psf_library_x_dim = seed.psf_library.data.shape[2] / seed.psf_library.oversampling
    seed.psf_library_y_dim = seed.psf_library.data.shape[1] / seed.psf_library.oversampling

    # Check the wrapped function, but call using 'aperture' rather than
    # 'full_frame'
    expected_k1l1 = [(850, 1151, 850, 1151, 0, 301, 0, 301),
                     (0, 201, 250, 551, 0, 201, 0, 301),
                     (850, 1151, 0, 176, 0, 301, 0, 176),
                     (1750, 2048, 1850, 2048, 0, 298, 0, 198),
                     (0, 150, 0, 150, 0, 150, 0, 150)]

    seed.coord_adjust['xoffset'] = 0  # Already defined, but let's be explicit
    seed.coord_adjust['yoffset'] = 0  # Already defined, but let's be explicit
    seed.output_dims = [2048, 2048]
    seed.psf_library_core_x_dim = 301
    seed.psf_library_core_y_dim = 301
    seed.params = {'simSignals': {}}
    seed.add_psf_wings = False
    seed.psf_library_oversamp = 1

    for index in range(5):
        psf, min_x, min_y, add_wings = seed.create_psf_stamp(tab[index]['pixelx'], tab[index]['pixely'],
                                                             stamp_dims[1], stamp_dims[0],
                                                             ignore_detector=False)
        stamp_x_loc = 301 // 2 - min_x
        stamp_y_loc = 301 // 2 - min_y
        updated_psf_dimensions = psf.shape

        xap, yap, xpts, ypts, (i1, i2), (j1, j2), (k1, k2), \
            (l1, l2) = seed.create_psf_stamp_coords(tab[index]['pixelx'], tab[index]['pixely'],
                                                    updated_psf_dimensions, stamp_x_loc, stamp_y_loc,
                                                    coord_sys='aperture')
        assert (i1, i2, j1, j2, k1, k2, l1, l2) == expected_k1l1[index]


def test_overlap_coordinates_subarray():
    """Test the function that calculates the coordinates for the
    overlap between two stamp images
    """
    seed = catalog_seed_image.Catalog_seed(offline=True)

    # Create a point sources table
    tab = Table()
    tab['index'] = np.arange(5).astype(np.int) + 1
    tab['pixelx'] = [200, 50, 200, 350, -0.5]
    tab['pixely'] = [200, 200, 25, 350, -0.5]
    tab['RA_degrees'] = [12.008729, 11.999914, 12.003422, 11.995729, 11.995729]
    tab['Dec_degrees'] = [-0.00885006, 1.7231344e-09, -1.5512744e-05, -4.9949034e-05, -4.9949034e-05]
    tab['magnitude'] = np.repeat(14, 5)
    tab['countrate_e/s'] = np.repeat(161630.72, 5)
    tab['counts_per_frame_e'] = np.repeat(1735391.9, 5)

    subarray_expected_results = []
    subarray_expected_results.append((50, 351, 50, 351, 0, 301, 0, 301))
    subarray_expected_results.append((0, 201, 50, 351, 100, 301, 0, 301))
    subarray_expected_results.append((50, 351, 0, 176, 0, 301, 125, 301))
    subarray_expected_results.append((200, 400, 200, 400, 0, 200, 0, 200))
    subarray_expected_results.append((0, 150, 0, 150, 151, 301, 151, 301))

    # Test the basic function using a subarray
    stamp_dims = (301, 301)
    stamp_x = 150
    stamp_y = 150
    aperture_dims = (400, 400)  # Assume 400x400 pixel subarray
    for index in range(5):
        results = seed.cropped_coords(tab[index]['pixelx'], tab[index]['pixely'], aperture_dims, stamp_x,
                                      stamp_y, stamp_dims, ignore_detector=False)
    assert results == subarray_expected_results[index]

    # Test the wrapper function
    seed.coord_adjust['xoffset'] = 0  # Already defined, but let's be explicit
    seed.coord_adjust['yoffset'] = 0  # Already defined, but let's be explicit
    seed.output_dims = [400, 400]  # Assume 400x400 pixel subarray
    seed.subarray_bounds = [1648, 1648, 2047, 2047]
    seed.ffsize = 2048
    seed.psf_library_core_x_dim = 301
    seed.psf_library_core_y_dim = 301
    seed.params = {'simSignals': {}}
    seed.add_psf_wings = False
    seed.psf_library_oversamp = 1
    for index in range(5):
        results = seed.create_psf_stamp_coords(tab[index]['pixelx'], tab[index]['pixely'], stamp_dims,
                                               stamp_x, stamp_y, coord_sys='aperture', ignore_detector=False)
        ijkl = []
        for ele in results[4:]:
            ijkl.extend(list(ele))
        ijkl = tuple(ijkl)
    assert ijkl == subarray_expected_results[index]

    # Dummy PSF library
    seed.psf_library = create_dummy_psf_grid()
    seed.psf_library_x_dim = seed.psf_library.data.shape[2] / seed.psf_library.oversampling
    seed.psf_library_y_dim = seed.psf_library.data.shape[1] / seed.psf_library.oversampling

    # Test the wrapper function using a modified PSF stamp image shape
    expected_k1l1 = [(50, 351, 50, 351, 0, 301, 0, 301),
                     (0, 201, 50, 351, 100, 301, 0, 301),
                     (50, 351, 0, 176, 0, 301, 125, 301),
                     (200, 400, 200, 400, 0, 200, 0, 200),
                     (0, 150, 0, 150, 151, 301, 151, 301)]
    for index in range(5):
        psf, min_x, min_y, add_wings = seed.create_psf_stamp(tab[index]['pixelx'], tab[index]['pixely'],
                                                             stamp_dims[1], stamp_dims[0],
                                                             ignore_detector=False)
        stamp_x_loc = 301 // 2 - min_x
        stamp_y_loc = 301 // 2 - min_y
        updated_psf_dimensions = psf.shape

        xap, yap, xpts, ypts, (i1, i2), (j1, j2), (k1, k2), \
            (l1, l2) = seed.create_psf_stamp_coords(tab[index]['pixelx'], tab[index]['pixely'],
                                                    updated_psf_dimensions, stamp_x_loc, stamp_y_loc,
                                                    coord_sys='aperture')
        assert (i1, i2, j1, j2, k1, k2, l1, l2) == expected_k1l1[index]



#! /usr/bin/env python

"""This module contains code to make, locate, and parse the appropriate segment
PSF library files to use for a given simulation. These files are made using
the WebbPSF psf_grid() method, and are turned into photutils GriddedPSFModel
objects using webbpsf.utils.to_griddedpsfmodel.

Author
------

    - Lauren Chambers

Use
---

    This module can be imported and called as such:
    ::
        from mirage.psf.segment_psfs import get_gridded_segment_psf_library_list
        lib = get_gridded_segment_psf_library_list(instrument, detector, filter,
                out_dir, pupilname="CLEAR")
"""
import os
import time

from astropy.io import fits
import numpy as np
import pysiaf
import webbpsf
from webbpsf.gridded_library import CreatePSFLibrary
from webbpsf.utils import to_griddedpsfmodel

from mirage.psf.psf_selection import get_library_file


def generate_segment_psfs(ote, segment_tilts, out_dir, filters=['F212N', 'F480M'],
                          detectors='all', fov_pixels=1024, overwrite=False):
    """Generate NIRCam PSF libraries for all 18 mirror segments given a perturbed OTE
    mirror state. Saves each PSF library as a FITS file named in the following format:
        nircam_{filter}_fovp{fov size}_samp1_npsf1_seg{segment number}.fits

    Parameters
    ----------
    ote : webbpsf.opds.OTE_Linear_Model_WSS object
        WebbPSF OTE object describing perturbed OTE state with tip and tilt removed

    segment_tilts : numpy.ndarray
        List of X and Y tilts for each mirror segment, in microradians

    out_dir : str
        Directory in which to save FITS files

    filters : str or list, optional
        Which filters to generate PSF libraries for. Default is ['F212N', 'F480M']
        (the two filters used for most commissioning activities).

    detectors : str or list, optional
        Which detectors to generate PSF libraries for. Default is 'all'.

    fov_pixels : int, optional
        Size of the PSF to generate, in pixels. Default is 1024.

    overwrite : bool, optional
            True/False boolean to overwrite the output file if it already
            exists. Default is True.

    """
    # Create webbpsf NIRCam instance
    nc = webbpsf.NIRCam()

    # Create dummy CreatePSFLibrary instance to get lists of filter and detectors
    lib = CreatePSFLibrary

    # Define the filter list to loop through
    if isinstance(filters, str):
        filters = [filters]
    elif not isinstance(filters, list):
        raise TypeError('Please define filters as a string or list, not {}'.format(type(filters)))

    # Define the detector list to loop through
    if detectors == 'all':
        detectors = ['NRCA1', 'NRCA2', 'NRCA3', 'NRCA4', 'NRCA5',
                     'NRCB1', 'NRCB2', 'NRCB3', 'NRCB4', 'NRCB5',]
    elif isinstance(detectors, str):
        detectors = [detectors]
    elif not isinstance(detectors, list):
        raise TypeError('Please define detectors as a string or list, not {}'.format(type(detectors)))


    # Create PSF grids for all segments, detectors, and filters
    for i in range(18):
        start_time = time.time()
        i_segment = i + 1
        segname = webbpsf.webbpsf_core.segname(i_segment)
        print('GENERATING SEGMENT {} DATA'.format(segname))
        print('------------------------------')

        det_filt_match = False
        for det in sorted(detectors):
            for filt in list(filters):
                # Make sure the detectors and filters match
                if (det in lib.nrca_short_detectors and filt not in lib.nrca_short_filters) \
                        or (det in lib.nrca_long_detectors and filt not in lib.nrca_long_filters):
                    continue

                det_filt_match = True

                # Define the filter and detector
                nc.filter = filt
                nc.detector = det

                # Restrict the pupil to the current segment
                pupil = webbpsf.webbpsf_core.one_segment_pupil(i_segment)
                ote.amplitude = pupil[0].data
                nc.pupil = ote

                # Generate the PSF grid
                # NOTE: we are choosing a polychromatic simulation here to better represent the
                # complexity of simulating unstacked PSFs. See the WebbPSF website for more details.
                grid = nc.psf_grid(num_psfs=1, save=False, all_detectors=False,
                                   use_detsampled_psf=True, fov_pixels=fov_pixels,
                                   oversample=1, overwrite=overwrite, add_distortion=False,
                                   nlambda=10)

                # Remove and add header keywords about segment
                del grid.meta["grid_xypos"]
                del grid.meta["oversampling"]
                grid.meta['SEGID'] = (i_segment, 'ID of the mirror segment')
                grid.meta['SEGNAME'] = (segname, 'Name of the mirror segment')
                grid.meta['XTILT'] = (round(segment_tilts[i, 0], 2), 'X tilt of the segment in microns')
                grid.meta['YTILT'] = (round(segment_tilts[i, 1], 2), 'Y tilt of the segment in microns')

                # Write out file
                filename = 'nircam_{}_{}_fovp{}_samp1_npsf1_seg{:02d}.fits'.format(det.lower(), filt.lower(),
                                                                                   fov_pixels, i_segment)
                filepath = os.path.join(out_dir, filename)
                primaryhdu = fits.PrimaryHDU(grid.data)
                tuples = [(a, b, c) for (a, (b, c)) in sorted(grid.meta.items())]
                primaryhdu.header.extend(tuples)
                hdu = fits.HDUList(primaryhdu)
                hdu.writeto(filepath, overwrite=overwrite)
                print('Saved gridded library file to {}'.format(filepath))

        if det_filt_match == False:
            raise ValueError('No matching filters and detectors given - all '
                             'filters are longwave but detectors are shortwave, '
                             'or vice versa.')

        print('\nElapsed time:', time.time() - start_time, '\n')


def get_gridded_segment_psf_library_list(instrument, detector, filtername,
                                         library_path, pupilname="CLEAR"):
    """Find the filenames for the appropriate gridded segment PSF libraries and
    read them into griddedPSFModel objects

    Parameters
    ----------
    instrument : str
        Name of instrument the PSFs are from

    detector : str
        Name of the detector within ```instrument```

    filtername : str
        Name of filter used for PSF library creation

    library_path : str
        Path pointing to the location of the PSF library

    pupilname : str, optional
        Name of pupil wheel element used for PSF library creation. Default is "CLEAR".

    Returns:
    --------
    libraries : list of photutils.griddedPSFModel
        List of object containing segment PSF libraries

    """
    library_list = get_segment_library_list(instrument, detector, filtername, library_path, pupil=pupilname)

    print("Segment PSFs will be generated using:")
    for filename in library_list:
        print(os.path.basename(filename))

    libraries = []
    for filename in library_list:
        with fits.open(filename) as hdulist:
        #     hdr = hdulist[0].header
        #     d = hdulist[0].data
        #
        # data = d[0][0]
        # phdu = fits.PrimaryHDU(data, header=hdr)
        # hdulist = fits.HDUList(phdu)

            lib_model = to_griddedpsfmodel(hdulist)
            libraries.append(lib_model)

    return libraries


def get_segment_library_list(instrument, detector, filt,
                             library_path, pupil='CLEAR'):
    """Given an instrument and filter name along with the path of
    the PSF library, find the appropriate 18 segment PSF library files.

    Parameters
    -----------
    instrument : str
        Name of instrument the PSFs are from

    detector : str
        Name of the detector within ```instrument```

    filt : str
        Name of filter used for PSF library creation

    library_path : str
        Path pointing to the location of the PSF library

    pupil : str, optional
        Name of pupil wheel element used for PSF library creation. Default is
        'CLEAR'.

    segment_id : int or None, optional
        If specified, returns a segment PSF library file and denotes the ID
        of the mirror segment

    Returns
    --------
    library_list : list
        List of the names of the segment PSF library files for the instrument
        and filter name
    """
    library_list = []
    for seg_id in np.arange(1, 19):
         segment_file = get_library_file(
             instrument, detector, filt, pupil, '', 0, library_path,
             segment_id=seg_id
         )
         library_list.append(segment_file)

    return library_list


def get_segment_offset(segment_number, detector, library_list):
    """Convert vectors coordinates in the local segment control
    coordinate system to NIRCam detector X and Y coordinates,
    at least proportionally, in order to calculate the location
    of the segment PSFs on the given detector.

    Parameters
    ----------
    segment : int
        Segment ID, i.e 3
    detector : str
        Name of NIRCam detector
    library_list : list
        List of the names of the segment PSF library files

    Returns
    -------
    x_displacement
        The shift of the segment PSF in NIRCam SW x pixels
    y_displacement
        The shift of the segment PSF in NIRCam SW y pixels
    """

    # Verify that the segment number in the header matches the index
    seg_index = int(segment_number) - 1
    header = fits.getheader(library_list[seg_index])

    assert int(header['SEGID']) == int(segment_number), \
        "Uh-oh. The segment ID of the library does not match the requested " \
        "segment. The library_list was not assembled correctly."
    xtilt = header['XTILT']
    ytilt = header['YTILT']
    segment = header['SEGNAME'][:2]

    # These conversion factors were empirically calculated by measuring the
    # relation between tilt and the pixel displacement
    tilt_to_pixel_slope = 13.4
    tilt_to_pixel_intercept = 0

    control_xaxis_rotations = {
        'A1': 180, 'A2': 120, 'A3': 60, 'A4': 0, 'A5': -60,
        'A6': -120, 'B1': 0, 'C1': 60, 'B2': -60, 'C2': 0,
        'B3': -120, 'C3': -60, 'B4': -180, 'C4': -120,
        'B5': -240, 'C5': -180, 'B6': -300, 'C6': -240
    }

    x_rot = control_xaxis_rotations[segment]  # degrees
    x_rot_rad = x_rot * np.pi / 180  # radians

    # Note that y is defined as the x component and x is defined as the y component.
    # This is because "xtilt" moves the PSF in the y direction, and vice versa.
    tilt_onto_y = (xtilt * np.cos(x_rot_rad)) - (ytilt * np.sin(x_rot_rad))
    tilt_onto_x = (xtilt * np.sin(x_rot_rad)) + (ytilt * np.cos(x_rot_rad))

    # TODO: IS THE SLOPE DIFFERENT FOR LW DETECTORS????
    x_displacement = -(tilt_onto_x * tilt_to_pixel_slope) + tilt_to_pixel_intercept  # pixels
    y_displacement = -(tilt_onto_y * tilt_to_pixel_slope) + tilt_to_pixel_intercept  # pixels

    # Get the appropriate pixel scale from pysiaf
    siaf = pysiaf.Siaf('nircam')
    aperture = siaf['NRC{}_FULL'.format(detector[-2:].upper())]
    nircam_x_pixel_scale = aperture.XSciScale  # arcsec/pixel
    nircam_y_pixel_scale = aperture.YSciScale  # arcsec/pixel

    # Convert the pixel displacement into angle
    x_arcsec = x_displacement * nircam_x_pixel_scale  # arcsec
    y_arcsec = y_displacement * nircam_y_pixel_scale  # arcsec

    return x_arcsec, y_arcsec



"""Constant parameters in support of mirage simulations."""

# NIRISS filter wheel elements
# Source: CSA-JWST-CD-0003: Near Infrared Imager and Slitless Spectrograph Operations Concept Document
NIRISS_PUPIL_WHEEL_ELEMENTS = 'CLEARP F090W F115W F140M F150W F158M F200W NRM GR700XD'.split()
NIRISS_FILTER_WHEEL_ELEMENTS = 'CLEAR GR150R GR150C F277W F356W F380M F430M F444W F480M'.split()


'''Define unit tests that cover the addition of a TSO source to a seed image

Authors
-------
    - Bryan Hilbert

Use
---
    Ensure you have pytest installed. Then, simply run pytest in any
    parent directory of mirage/tests/:
    >>> pytest
'''
import astropy.units as u
import numpy as np
from synphot import units

from mirage.seed_image import segmentation_map as segmap
from mirage.seed_image import tso
from mirage.utils.constants import FLAMBDA_CGS_UNITS

def test_add_tso_sources():
    """Test the addition of the TSO source to the seed image
    """
    background_seed_image = np.zeros((10, 10))
    background_seed_image[1:4, 1:4] = 100

    background_seg_map = np.zeros((10, 10)).astype(np.int)
    background_seg_map[1:4, 1:4] = 2

    tso_seed = np.zeros((10, 10))
    tso_seed[3:6, 3:6] = 1000
    tso_seeds = [tso_seed]

    tso_seg = segmap.SegMap()
    tso_seg.ydim, tso_seg.xdim = (10, 10)
    tso_seg.initialize_map()
    tso_seg.segmap[3:6, 3:6] = 99999
    tso_segs = [tso_seg]

    light_curve = {}
    light_curve['times'] = np.arange(0, 10) * u.second
    light_curve['fluxes'] = np.repeat(1., len(light_curve['times'])) * FLAMBDA_CGS_UNITS
    lcs = [light_curve]

    frame_time = 2.
    total_frames = 5
    exposure_frames = 5
    frames_per_int = 5
    num_ints = 1
    resets = 1

    seed, seg = tso.add_tso_sources(background_seed_image, background_seg_map, tso_seeds, tso_segs, lcs, frame_time,
                                    total_frames, exposure_frames, frames_per_int, num_ints, resets,
                                    starting_time=0, starting_frame=0, samples_per_frametime=5)

    # Make sure the seed image is 4D and the segmentation map is 2D
    assert len(seed.shape) == 4
    assert len(seg.shape) == 2

    # Check the signal levels in the seed image frames
    assert np.all(seed[0, :, 2, 2] == np.array([200., 400., 600., 800., 1000.]))
    assert np.all(seed[0, :, 5, 5] == np.array([2000., 4000., 6000., 8000., 10000.]))

    # Check that both objects are in the segmentation map
    assert np.unique(seg[1:3, 1:3]) == [2]
    assert np.unique(seg[3:6, 3:6]) == [99999]


def test_check_lightcurve_time():
    """Check that the lightcurve time is compared to the total
    exposure time correctly
    """
    exposure_time = 1000.
    frame_time = 1.

    # Case where light curve is too short
    light_curve = {}
    light_curve['times'] = np.arange(0, 500) * u.second
    light_curve['fluxes'] = np.repeat(1., len(light_curve['times'])) * FLAMBDA_CGS_UNITS
    new_lc = tso.check_lightcurve_time(light_curve, exposure_time, frame_time)
    assert np.max(new_lc['times'].data) >= exposure_time

    # Case where lightcurve start time is < 0
    light_curve = {}
    light_curve['times'] = np.arange(-10, 500) * u.second
    light_curve['fluxes'] = np.repeat(1., len(light_curve['times'])) * FLAMBDA_CGS_UNITS
    new_lc = tso.check_lightcurve_time(light_curve, exposure_time, frame_time)
    assert np.min(new_lc['times'].data) == 0.

    # Case where lightcurve start time is > 0
    light_curve = {}
    light_curve['times'] = np.arange(30, 500) * u.second
    light_curve['fluxes'] = np.repeat(1., len(light_curve['times'])) * FLAMBDA_CGS_UNITS
    new_lc = tso.check_lightcurve_time(light_curve, exposure_time, frame_time)
    assert np.min(new_lc['times'].data) == 0.


def test_interpolate_lightcurve():
    """Test that a lightcurve is correctly interpolated to have the
    correct number of entries
    """
    samples_per_frame_time = 5
    frame_time = 1.

    # Create simple lightcurve
    light_curve = {}
    light_curve['times'] = np.arange(0, 10) * u.second
    light_curve['fluxes'] = np.repeat(1., len(light_curve['times'])) * FLAMBDA_CGS_UNITS

    # Interpolate
    interp = tso.interpolate_lightcurve(light_curve, samples_per_frame_time, frame_time)

    assert len(interp['times'].value) == (10 - 1) * frame_time * (samples_per_frame_time - 1)
    assert np.all(interp['times'].value[0:5] == [0, 0.25, 0.5, 0.75, 1.0])


def test_update_segmentation_map():
    """Test that the segmentation map is updated correctly when adding a
    new object
    """
    seg_map = np.zeros((10, 10))
    seg_map[3:6, 3:6] = 10

    new_obj = np.zeros((10, 10))
    new_obj[5:8, 5:8] = 999

    new_map = tso.update_segmentation_map(seg_map, new_obj)

    assert np.unique(new_map[3:5, 3:5]) == [10]
    assert np.unique(new_map[5:8, 5:8]) == [999]




"""Various universal constants and values that are useful across the MIRAGE package.

Use
---
    ::
        from mirage.utils import constants
        inst_abb = constants.instrument_abbreviations
"""
import astropy.units as u
import numpy as np

instrument_abbreviations = {'nircam': 'NRC',
                            'fgs': 'FGS',
                            'niriss': 'NIS',
                            'nirspec': 'NRS',
                            'miri': 'MIR'}

EXPTYPES = {"nircam": {"imaging": "NRC_IMAGE", "ts_imaging": "NRC_TSIMAGE",
                       "wfss": "NRC_WFSS", "ts_grism": "NRC_TSGRISM"},
            "niriss": {"imaging": "NIS_IMAGE", "ami": "NIS_IMAGE", "pom": "NIS_IMAGE",
                       "wfss": "NIS_WFSS"},
            "fgs": {"imaging": "FGS_IMAGE"}}

NIRISS_FILTER_WHEEL_FILTERS = ['F277W', 'F356W', 'F380M', 'F430M', 'F444W', 'F480M']
NIRISS_PUPIL_WHEEL_FILTERS = ['F090W', 'F115W', 'F158M', 'F140M', 'F150W', 'F200W']
NIRCAM_PUPIL_WHEEL_FILTERS = ['F162M', 'F164N', 'F323N', 'F405N', 'F466N', 'F470N']

FLAMBDA_CGS_UNITS = u.erg / u.second / u.cm / u.cm / u.AA
FLAMBDA_MKS_UNITS = u.watt / u.meter**2 / u.micron
FNU_CGS_UNITS = u.erg / u.second / u.cm / u.cm / u.Hz
FNU_MKS_UNITS = u.watt / u.m**2 / u.Hz

CATALOG_YAML_ENTRIES = ['pointsource', 'galaxyListFile', 'extended', 'movingTargetList',
                        'movingTargetSersic', 'movingTargetExtended', 'movingTargetToTrack',
                        'tso_imaging_catalog', 'tso_grism_catalog']

TSO_MODES = ['ts_imaging', 'ts_grism']

# Upper limit to the size of a seed image or dark current array. Arrays
# containing more pixels than this limit will be split into segment files.
FILE_SPLITTING_LIMIT = 38. * 2048. * 2048

CRDS_FILE_TYPES = {'badpixmask': 'mask',
                   'astrometric': 'distortion',
                   'gain': 'gain',
                   'ipc': 'ipc',
                   'linearity': 'linearity',
                   'pixelAreaMap': 'area',
                   'saturation': 'saturation',
                   'superbias': 'superbias',
                   'pixelflat': 'flat'}

MEAN_GAIN_VALUES = {'nircam': {'swa': 2.443376, 'swb': 2.4908085, 'lwa': 2.1923525, 'lwb': 2.1811192,
                               'nrca1': 2.4926066, 'nrca2': 2.5222762, 'nrca3': 2.4837794, 'nrca4': 2.2748423,
                               'nrcb1': 2.5201690, 'nrcb2': 2.5557644, 'nrcb3': 2.5512073, 'nrcb4': 2.3360932,
                               'nrca5': 2.1923525, 'nrcb5': 2.1811192},
                    'niriss': 1.611902,
                    'fgs': {'guider1': 1.9712874, 'guider2': 1.6897644}
                    }

# Factor by which throughput is decreased when the Grism is used compared
# to when it is not in the beam
NIRISS_GRISM_THROUGHPUT_FACTOR = 0.8

# For use in converting background MJy/sr to e-/sec
PRIMARY_MIRROR_AREA = 25.326 * u.meter * u.meter
PLANCK = 6.62607004e-34  * u.meter * u.meter * u.kg / u.second


def grism_factor(instrument_name):
    """Return the factor by which the field of view is expanded when
    creating grism simulations compared to direct image simulations

    Parameters
    ----------
    instrument_name : str
        JWST instrument name

    Returns
    -------
    factor : float
        Multiplicative factor by which the fov is enlarged
    """
    if instrument_name.lower() == 'nircam':
        return np.sqrt(2.)
    elif instrument_name.lower() == 'niriss':
        return 2322./2048.



"""Test the functions provided by mirage.yaml.yaml_generator

Authors
-------
    - Lauren Chambers

Use
---
    >>> pytest test_yaml_generator.py
"""
import os

import numpy as np
import pytest

from mirage.yaml.yaml_generator import SimInput

# Define directory and file locations
__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))

# Reset the MIRAGE_DATA env variable to be a real location so yaml_generator
# doesn't croak
# Determine if tests are being run on Travis
ON_TRAVIS = 'travis' in os.path.expanduser('~')
if not ON_TRAVIS:
    orig_mirage_data = os.environ['MIRAGE_DATA']

os.environ["MIRAGE_DATA"] = __location__
os.environ["CRDS_PATH"] = os.path.join(__location__, "temp")


def call_with_assorted_user_inputs(catalogs, crs, dates, backgrounds, roll_angles):
    """Test various user-inputs to yaml_generator, including dates, backgrounds,
    roll angles, cosmic rays, and catalogs
    """
    # Make an instance of the SimInput class
    input_xml = os.path.join(__location__, 'test_data/misc/54321/54321_niriss_wfss_prime_nircam_imaging_parallel_mult_targets.xml')
    pointing_file = os.path.join(__location__, 'test_data/misc/54321/54321_niriss_wfss_prime_nircam_imaging_parallel_mult_targets.pointing')
    temp_output_dir = os.path.join(__location__, "temp")

    yam = SimInput(input_xml, pointing_file, verbose=True, catalogs=catalogs,
                   offline=True, output_dir=temp_output_dir,
                   simdata_output_dir=temp_output_dir,
                   datatype='raw', reffile_defaults='crds',
                   cosmic_rays=crs, dates=dates, background=backgrounds,
                   roll_angle=roll_angles)
    yam.use_linearized_darks = True
    yam.create_inputs()
    return yam.info


def test_user_inputs_basic():
    """Check that use-input dates, roll angles, cosmic ray info,
    backgrounds, and catalogs are propagated correctly in cases
    with the more basic user inputs.
    """
    cats = {'TARG1': {'point_source': 'ptsrc1.cat',
                      'galaxy': 'galaxy1.cat',
                      'extended': 'ex1.cat',
                      'moving_pointsource': 'mt_ptsrc1.cat',
                      'moving_sersic': 'mt_gal_1.cat',
                      'moving_extended': 'mt_ext_1.cat',
                      'moving_target_to_track': 'mt_track_1.cat'
                      },
            'TARG2': {'point_source': 'ptsrc2.cat',
                      'galaxy': 'galaxy2.cat',
                      'extended': 'ex2.cat',
                      'moving_pointsource': 'mt_ptsrc2.cat',
                      'moving_sersic': 'mt_gal_2.cat',
                      'moving_extended': 'mt_ext_2.cat',
                      'moving_target_to_track': 'mt_track_2.cat'
                      }
            }
    cr = {'library': 'FLARE', 'scale': 44.0}
    date = '2019-5-25'
    background = {'001': 'high', '002': 'medium', '003': 22.3}
    roll_angle = 34.5

    tab = call_with_assorted_user_inputs(cats, cr, date, background, roll_angle)

    nrc_obs1 = ((np.array(tab['obs_num']) == '001') & (np.array(tab['Instrument']) == 'NIRCAM'))
    nis_obs1 = ((np.array(tab['obs_num']) == '001') & (np.array(tab['Instrument']) == 'NIRISS'))
    nrc_obs2 = ((np.array(tab['obs_num']) == '002') & (np.array(tab['Instrument']) == 'NIRCAM'))
    nis_obs2 = ((np.array(tab['obs_num']) == '002') & (np.array(tab['Instrument']) == 'NIRISS'))
    nrc_obs3 = ((np.array(tab['obs_num']) == '003') & (np.array(tab['Instrument']) == 'NIRCAM'))
    nis_obs3 = ((np.array(tab['obs_num']) == '003') & (np.array(tab['Instrument']) == 'NIRISS'))
    nrc_obs = np.vstack([nrc_obs1, nrc_obs2, nrc_obs3])
    nis_obs = np.vstack([nis_obs1, nis_obs2, nis_obs3])

    # Check dates
    date_match = [entry == date for entry in tab['Date']]
    assert all(date_match) is True

    # Check roll angle
    pav3_match = [entry == str(roll_angle) for entry in tab['PAV3']]
    assert all(pav3_match) is True

    # Check cosmic rays
    cr_lib_match = [entry == cr['library'] for entry in tab['CosmicRayLibrary']]
    assert all(cr_lib_match) is True

    cr_scale_match = [entry == str(cr['scale']) for entry in tab['CosmicRayScale']]
    assert all(cr_scale_match) is True

    # Check backgrounds
    for i, key in enumerate(background):
        back01_nis_match = [entry == str(background[key]) for entry in np.array(tab['BackgroundRate'])[nis_obs[i, :]]]
        assert all(back01_nis_match) is True
        back01_nrcsw_match = [entry == str(background[key]) for entry in np.array(tab['sw_bkgd'])[nrc_obs[i, :]]]
        assert all(back01_nrcsw_match) is True
        back01_nrclw_match = [entry == str(background[key]) for entry in np.array(tab['lw_bkgd'])[nrc_obs[i, :]]]
        assert all(back01_nrclw_match) is True

    # Check catalogs
    cwd = os.getcwd()

    for i, key in enumerate(['TARG1', 'TARG2', 'TARG1']):
        # NIRCAM
        cat01_nrc_ptsrc_match = [entry == os.path.join(cwd, cats[key]['point_source']) for entry in np.array(tab['sw_ptsrc'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_ptsrc_match) is True
        cat01_nrc_galaxy_match = [entry == os.path.join(cwd, cats[key]['galaxy']) for entry in np.array(tab['sw_galcat'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_galaxy_match) is True
        cat01_nrc_ext_match = [entry == os.path.join(cwd, cats[key]['extended']) for entry in np.array(tab['sw_ext'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_ext_match) is True
        cat01_nrc_mt_ptsrc_match = [entry == os.path.join(cwd, cats[key]['moving_pointsource']) for entry in np.array(tab['sw_movptsrc'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_mt_ptsrc_match) is True
        cat01_nrc_mt_gal_match = [entry == os.path.join(cwd, cats[key]['moving_sersic']) for entry in np.array(tab['sw_movgal'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_mt_gal_match) is True
        cat01_nrc_mt_ext_match = [entry == os.path.join(cwd, cats[key]['moving_extended']) for entry in np.array(tab['sw_movext'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_mt_ext_match) is True
        cat01_nrc_mt_track_match = [entry == os.path.join(cwd, cats[key]['moving_target_to_track']) for entry in np.array(tab['sw_solarsys'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_mt_track_match) is True

        # NIRISS
        cat01_nis_ptsrc_match = [entry == os.path.join(cwd, cats[key]['point_source']) for entry in np.array(tab['PointSourceCatalog'])[nis_obs[i, :]]]
        assert all(cat01_nis_ptsrc_match) is True
        cat01_nis_galaxy_match = [entry == os.path.join(cwd, cats[key]['galaxy']) for entry in np.array(tab['GalaxyCatalog'])[nis_obs[i, :]]]
        assert all(cat01_nis_galaxy_match) is True
        cat01_nis_ext_match = [entry == os.path.join(cwd, cats[key]['extended']) for entry in np.array(tab['ExtendedCatalog'])[nis_obs[i, :]]]
        assert all(cat01_nis_ext_match) is True
        cat01_nis_mt_ptsrc_match = [entry == os.path.join(cwd, cats[key]['moving_pointsource']) for entry in np.array(tab['MovingTargetList'])[nis_obs[i, :]]]
        assert all(cat01_nis_mt_ptsrc_match) is True
        cat01_nis_mt_gal_match = [entry == os.path.join(cwd, cats[key]['moving_sersic']) for entry in np.array(tab['MovingTargetSersic'])[nis_obs[i, :]]]
        assert all(cat01_nis_mt_gal_match) is True
        cat01_nis_mt_ext_match = [entry == os.path.join(cwd, cats[key]['moving_extended']) for entry in np.array(tab['MovingTargetExtended'])[nis_obs[i, :]]]
        assert all(cat01_nis_mt_ext_match) is True
        cat01_nis_mt_track_match = [entry == os.path.join(cwd, cats[key]['moving_target_to_track']) for entry in np.array(tab['MovingTargetToTrack'])[nis_obs[i, :]]]
        assert all(cat01_nis_mt_track_match) is True

    # Clean up
    temp_output_dir = os.path.join(__location__, "temp")
    os.system('rm -r {}'.format(temp_output_dir))


def test_user_inputs_complex():
    """Check that use-input dates, roll angles, cosmic ray info,
    backgrounds, and catalogs are propagated correctly in cases
    with the more complex user inputs.
    """
    cats = {'TARG1': {'nircam': {'point_source': 'ptsrc_nrc_1.cat',
                                 'galaxy': 'galaxy_nrc_1.cat',
                                 'extended': 'ex_nrc_1.cat',
                                 'moving_pointsource': 'mt_ptsrc_nrc_1.cat',
                                 'moving_sersic': 'mt_gal_nrc_1.cat',
                                 'moving_extended': 'mt_ext_nrc_1.cat',
                                 'moving_target_to_track': 'mt_track_nrc_1.cat'
                                 },
                      'niriss': {'point_source': 'ptsrc_nis_1.cat',
                                 'galaxy': 'galaxy_nis_1.cat',
                                 'extended': 'ex_nis_1.cat',
                                 'moving_pointsource': 'mt_ptsrc_nis_1.cat',
                                 'moving_sersic': 'mt_gal_nis_1.cat',
                                 'moving_extended': 'mt_ext_nis_1.cat',
                                 'moving_target_to_track': 'mt_track_nis_1.cat'
                                 }
                      },
            'TARG2': {'nircam': {'point_source': 'ptsrc_nrc_2.cat',
                                 'galaxy': 'galaxy_nrc_2.cat',
                                 'extended': 'ex_nrc_2.cat',
                                 'moving_pointsource': 'mt_ptsrc_nrc_2.cat',
                                 'moving_sersic': 'mt_gal_nrc_2.cat',
                                 'moving_extended': 'mt_ext_nrc_2.cat',
                                 'moving_target_to_track': 'mt_track_nrc_2.cat'
                                 },
                      'niriss': {'point_source': 'ptsrc_nis_2.cat',
                                 'galaxy': 'galaxy_nis_2.cat',
                                 'extended': 'ex_nis_2.cat',
                                 'moving_pointsource': 'mt_ptsrc_nis_2.cat',
                                 'moving_sersic': 'mt_gal_nis_2.cat',
                                 'moving_extended': 'mt_ext_nis_2.cat',
                                 'moving_target_to_track': 'mt_track_nis_2.cat'
                                 }
                      },
            }

    cr = {'001': {'library': 'FLARE', 'scale': 1.2},
          '002': {'library': 'SUNMIN', 'scale': 5.5},
          '003': {'library': 'SUNMAX', 'scale': 4.4}}

    date = {'001': '2019-05-25', '002': '2019-11-15', '003': '2020-10-14'}

    background = {'001': {'nircam': {'sw': 0.2, 'lw': 0.3}, 'niriss': 0.4},
                  '002': {'nircam': {'sw': 'medium', 'lw': 'high'}, 'niriss': 'low'},
                  '003': {'nircam': {'sw': 0.75, 'lw': 'high'}, 'niriss': 0.2}}

    roll_angle = {'001': 34.5, '002': 154.5, '003': 3.14}

    tab = call_with_assorted_user_inputs(cats, cr, date, background, roll_angle)

    nrc_obs1 = ((np.array(tab['obs_num']) == '001') & (np.array(tab['Instrument']) == 'NIRCAM'))
    nis_obs1 = ((np.array(tab['obs_num']) == '001') & (np.array(tab['Instrument']) == 'NIRISS'))
    nrc_obs2 = ((np.array(tab['obs_num']) == '002') & (np.array(tab['Instrument']) == 'NIRCAM'))
    nis_obs2 = ((np.array(tab['obs_num']) == '002') & (np.array(tab['Instrument']) == 'NIRISS'))
    nrc_obs3 = ((np.array(tab['obs_num']) == '003') & (np.array(tab['Instrument']) == 'NIRCAM'))
    nis_obs3 = ((np.array(tab['obs_num']) == '003') & (np.array(tab['Instrument']) == 'NIRISS'))
    nrc_obs = np.vstack([nrc_obs1, nrc_obs2, nrc_obs3])
    nis_obs = np.vstack([nis_obs1, nis_obs2, nis_obs3])

    # Check dates
    for i, key in enumerate(date):
        date_match = [entry == date[key] for entry in np.array(tab['Date'])[nis_obs[i, :]]]
        assert all(date_match) is True
        date_match = [entry == date[key] for entry in np.array(tab['Date'])[nrc_obs[i, :]]]
        assert all(date_match) is True

    # Check roll angle
    for i, key in enumerate(roll_angle):
        pav3_match = [entry == str(roll_angle[key]) for entry in np.array(tab['PAV3'])[nis_obs[i, :]]]
        assert all(pav3_match) is True
        pav3_match = [entry == str(roll_angle[key]) for entry in np.array(tab['PAV3'])[nrc_obs[i, :]]]
        assert all(pav3_match) is True

    # Check cosmic rays
    for i, key in enumerate(cr):
        cr_lib_match = [entry == cr[key]['library'] for entry in np.array(tab['CosmicRayLibrary'])[nis_obs[i, :]]]
        assert all(cr_lib_match) is True
        cr_lib_match = [entry == cr[key]['library'] for entry in np.array(tab['CosmicRayLibrary'])[nrc_obs[i, :]]]
        assert all(cr_lib_match) is True

        cr_scale_match = [entry == str(cr[key]['scale']) for entry in np.array(tab['CosmicRayScale'])[nis_obs[i, :]]]
        assert all(cr_scale_match) is True
        cr_scale_match = [entry == str(cr[key]['scale']) for entry in np.array(tab['CosmicRayScale'])[nrc_obs[i, :]]]
        assert all(cr_scale_match) is True

    # Check backgrounds
    for i, key in enumerate(background):
        back01_nis_match = [entry == str(background[key]['niriss']) for entry in np.array(tab['BackgroundRate'])[nis_obs[i, :]]]
        assert all(back01_nis_match) is True
        back01_nrcsw_match = [entry == str(background[key]['nircam']['sw']) for entry in np.array(tab['sw_bkgd'])[nrc_obs[i, :]]]
        assert all(back01_nrcsw_match) is True
        back01_nrclw_match = [entry == str(background[key]['nircam']['lw']) for entry in np.array(tab['lw_bkgd'])[nrc_obs[i, :]]]
        assert all(back01_nrclw_match) is True

    # Check catalogs
    cwd = os.getcwd()

    for i, key in enumerate(['TARG1', 'TARG2', 'TARG1']):
        # NIRCAM
        cat01_nrc_ptsrc_match = [entry == os.path.join(cwd, cats[key]['nircam']['point_source']) for entry in np.array(tab['sw_ptsrc'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_ptsrc_match) is True
        cat01_nrc_galaxy_match = [entry == os.path.join(cwd, cats[key]['nircam']['galaxy']) for entry in np.array(tab['sw_galcat'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_galaxy_match) is True
        cat01_nrc_ext_match = [entry == os.path.join(cwd, cats[key]['nircam']['extended']) for entry in np.array(tab['sw_ext'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_ext_match) is True
        cat01_nrc_mt_ptsrc_match = [entry == os.path.join(cwd, cats[key]['nircam']['moving_pointsource']) for entry in np.array(tab['sw_movptsrc'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_mt_ptsrc_match) is True
        cat01_nrc_mt_gal_match = [entry == os.path.join(cwd, cats[key]['nircam']['moving_sersic']) for entry in np.array(tab['sw_movgal'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_mt_gal_match) is True
        cat01_nrc_mt_ext_match = [entry == os.path.join(cwd, cats[key]['nircam']['moving_extended']) for entry in np.array(tab['sw_movext'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_mt_ext_match) is True
        cat01_nrc_mt_track_match = [entry == os.path.join(cwd, cats[key]['nircam']['moving_target_to_track']) for entry in np.array(tab['sw_solarsys'])[nrc_obs[i, :]]]
        assert all(cat01_nrc_mt_track_match) is True

        # NIRISS
        cat01_nis_ptsrc_match = [entry == os.path.join(cwd, cats[key]['niriss']['point_source']) for entry in np.array(tab['PointSourceCatalog'])[nis_obs[i, :]]]
        assert all(cat01_nis_ptsrc_match) is True
        cat01_nis_galaxy_match = [entry == os.path.join(cwd, cats[key]['niriss']['galaxy']) for entry in np.array(tab['GalaxyCatalog'])[nis_obs[i, :]]]
        assert all(cat01_nis_galaxy_match) is True
        cat01_nis_ext_match = [entry == os.path.join(cwd, cats[key]['niriss']['extended']) for entry in np.array(tab['ExtendedCatalog'])[nis_obs[i, :]]]
        assert all(cat01_nis_ext_match) is True
        cat01_nis_mt_ptsrc_match = [entry == os.path.join(cwd, cats[key]['niriss']['moving_pointsource']) for entry in np.array(tab['MovingTargetList'])[nis_obs[i, :]]]
        assert all(cat01_nis_mt_ptsrc_match) is True
        cat01_nis_mt_gal_match = [entry == os.path.join(cwd, cats[key]['niriss']['moving_sersic']) for entry in np.array(tab['MovingTargetSersic'])[nis_obs[i, :]]]
        assert all(cat01_nis_mt_gal_match) is True
        cat01_nis_mt_ext_match = [entry == os.path.join(cwd, cats[key]['niriss']['moving_extended']) for entry in np.array(tab['MovingTargetExtended'])[nis_obs[i, :]]]
        assert all(cat01_nis_mt_ext_match) is True
        cat01_nis_mt_track_match = [entry == os.path.join(cwd, cats[key]['niriss']['moving_target_to_track']) for entry in np.array(tab['MovingTargetToTrack'])[nis_obs[i, :]]]
        assert all(cat01_nis_mt_track_match) is True

    # Clean up
    temp_output_dir = os.path.join(__location__, "temp")
    os.system('rm -r {}'.format(temp_output_dir))


def test_get_psf_path():
    """Test that the get_psf_path method of the yaml_generator.SimInput class
    is working as expected.
    """
    # Make an instance of the SimInput class
    input_xml = os.path.join(__location__, 'test_data', 'NIRCam', '1144-OTE-10.xml')
    pointing_file = os.path.join(__location__, 'test_data', 'NIRCam', '1144-OTE-10.pointing')
    yam = SimInput(input_xml, pointing_file, offline=True)

    # Fake the activity IDs and instrument
    n_activities = 101
    act_ids = [np.base_repr(i, 36).zfill(2)  for i in range(n_activities)]
    yam.info = {}
    yam.info['act_id'] = act_ids
    yam.info['Instrument'] = ['NIRCam'] * n_activities

    # Test for a default path
    paths_out = yam.get_psf_path()
    assert len(paths_out) == n_activities,\
        'Default PSF path not properly provided.'
    assert 'nircam/gridded_psf_library' in paths_out[0],\
        'Default PSF path not properly provided.'
    np.testing.assert_array_equal(paths_out, paths_out[0],
                                  err_msg='Default PSF path not properly defined.')

    # Test for a single path
    yam.psf_paths = __location__
    paths_out = yam.get_psf_path()
    assert paths_out == [__location__] * n_activities,\
        'Single PSF path not properly assigned.'

    # Test for a list of paths of incorrect length
    with pytest.raises(ValueError) as e:
        yam.psf_paths = [__location__] * 99
        yam.get_psf_path()
        assert 'Please provide the psf_paths in the form of a list of strings ' \
               'with a length equal to the number of activities in the APT ' \
               'program (101), not equal to 99.' in e, \
            'Failed to reject psf_path of incorrect length.'

    # Test for a list of paths of correct length
    list_101_paths = [__location__] * 50 + [os.path.dirname(__location__)] * 51
    yam.psf_paths = list_101_paths
    paths_out = yam.get_psf_path()
    assert paths_out == sorted(list_101_paths),\
        'List of PSF paths not properly assigned.'

    # Test for a completely invalid path
    with pytest.raises(TypeError) as e:
        yam.psf_paths = 3.054
        yam.get_psf_path()
        assert 'Please provide the psf_paths in the form of a list or ' \
               'string, not float' in e, \
            'Failed to reject psf_path of incorrect type.'


def test_reffile_crds():
    """Make this a copy of test_reffile_crds_full_name (including reffile_overrides)
    but use 'crds'
    """
    catalogs = None

    # Make an instance of the SimInput class
    input_xml = os.path.join(__location__, 'test_data/misc/12345/12345_nircam_imaging_prime_niriss_wfss_parallel.xml')
    pointing_file = os.path.join(__location__, 'test_data/misc/12345/12345_nircam_imaging_prime_niriss_wfss_parallel.pointing')
    temp_output_dir = os.path.join(__location__, "temp")

    yam = SimInput(input_xml, pointing_file, verbose=True, catalogs=catalogs,
                   offline=True, output_dir=temp_output_dir,
                   simdata_output_dir=temp_output_dir,
                   datatype='raw', reffile_defaults='crds')
    yam.use_linearized_darks = True
    yam.create_inputs()

    # Check that all reference files are simply the string 'crds'
    expected_value = 'crds'
    assert np.all(np.array(yam.info['superbias']) == expected_value)
    assert np.all(np.array(yam.info['linearity']) == expected_value)
    assert np.all(np.array(yam.info['saturation']) == expected_value)
    assert np.all(np.array(yam.info['gain']) == expected_value)
    assert np.all(np.array(yam.info['astrometric']) == expected_value)
    assert np.all(np.array(yam.info['pixelAreaMap']) == expected_value)
    assert np.all(np.array(yam.info['badpixmask']) == expected_value)
    assert np.all(np.array(yam.info['ipc']) == expected_value)

# Clean up
    os.system('rm -r {}'.format(temp_output_dir))


def test_reffile_crds_full_name():
    """Test that the correct values for reference files are found
    """
    reffile_overrides = {'nircam': {'superbias': {'nrcb5': {'bright1': 'my_reffiles/my_superbias_for_b5.fits',
                                                            'shallow4': 'my_reffiles/my_superbias_for_b5.fits'
                                                            },
                                                  'nrcb4': {'shallow2': 'my_reffiles/my_superbias_for_b4.fits'}
                                                  },
                                    'linearity': {'nrcb5': 'my_reffiles/my_linearity_for_b5.fits',
                                                  'nrcb4': 'my_reffiles/my_linearity_for_b4.fits'},
                                    'saturation': {'nrcb5': 'my_reffiles/my_saturation_for_b5.fits',
                                                   'nrcb4': 'my_reffiles/my_saturation_for_b4.fits'},
                                    'gain': {'nrcb5': 'my_reffiles/my_gain_for_b5.fits',
                                             'nrcb4': 'my_reffiles/my_gain_for_b4.fits'},
                                    'distortion': {'nrcb5': {'f322w2': {'NRC_IMAGE': 'my_reffiles/my_distortion_for_b5.asdf'}},
                                                   'nrcb4': {'f150w': {'NRC_IMAGE': 'my_reffiles/my_distortion_for_b4.asdf'}}},
                                    'area': {'nrcb5': {'f322w2': {'clear': {'nrc_image': 'my_reffiles/my_pam_for_b5.fits'}}},
                                             'nrcb4': {'f150w': {'clear': {'nrc_image': 'my_reffiles/my_pam_for_b4.fits'}}}},
                                    'badpixmask': {'nrcb5': 'my_reffiles/my_bpm_for_b5.fits',
                                                   'nrcb4': 'my_reffiles/my_bpm_for_b4.fits'},
                                    'pixelflat': {'nrcb5': {'f322w2': {'clear': 'my_reffiles/my_flatfield_for_b5.fits'}}}
                                    },
                         'niriss': {'superbias': {'nisrapid': 'my_niriss_supebias.fits'},
                                    'linearity': 'my_niriss_linearity,fits',
                                    'saturation': 'my_niriss_saturation.fits',
                                    'gain': 'my_niriss_gain.fits',
                                    'distortion': {'F115W': {'nis_image': 'my_niriss_disotrtion.asdf'}},
                                    'area': {'clear': {'f115w': {'nis_image': 'my_niriss_area.fits'}}},
                                    'badpixmask': 'my_niriss_badpixmask.fits',
                                    'pixelflat': {'clear': {'f115w': 'my_niriss_flatfield.fits'}}
                                    }
                         }

    catalogs = None

    # Make an instance of the SimInput class
    input_xml = os.path.join(__location__, 'test_data/misc/12345/12345_nircam_imaging_prime_niriss_wfss_parallel.xml')
    pointing_file = os.path.join(__location__, 'test_data/misc/12345/12345_nircam_imaging_prime_niriss_wfss_parallel.pointing')
    temp_output_dir = os.path.join(__location__, "temp")

    yam = SimInput(input_xml, pointing_file, verbose=True, catalogs=catalogs,
                   offline=True, output_dir=temp_output_dir,
                   simdata_output_dir=temp_output_dir,
                   datatype='raw', reffile_defaults='crds_full_name',
                   reffile_overrides=reffile_overrides)
    yam.use_linearized_darks = True
    yam.create_inputs()

    # Make into numpy arrays for easier matching
    sw_filternames = np.array(yam.info['ShortFilter'])
    lw_filternames = np.array(yam.info['LongFilter'])
    sw_pupilnames = np.array(yam.info['ShortPupil'])
    lw_pupilnames = np.array(yam.info['LongPupil'])
    filternames = np.array(yam.info['FilterWheel'])
    pupilnames = np.array(yam.info['PupilWheel'])
    detectors = np.array(yam.info['detector'])
    instruments = np.array(yam.info['Instrument'])
    read_patterns = np.array(yam.info['ReadoutPattern'])

    match_nrc_sw_superbias = np.where((read_patterns == 'SHALLOW2') & (detectors == 'B4'))[0]
    match_nrc_lw_superbias = np.where((read_patterns == 'BRIGHT1') & (detectors == 'B5'))[0]
    match_nrc_sw_common = np.where(detectors == 'B4')[0]
    match_nrc_lw_common = np.where(detectors == 'B5')[0]
    match_nrc_sw_distortion_area = np.where((sw_filternames == 'F150W') & (detectors == 'B4') & (sw_pupilnames == 'CLEAR'))[0]
    match_nrc_lw_distortion_area = np.where((lw_filternames == 'F322W2') & (detectors == 'B5') & (lw_pupilnames == 'CLEAR'))[0]
    match_nrc_lw_flat = np.where((lw_filternames == 'F322W2') & (detectors == 'B5') & (lw_pupilnames == 'CLEAR'))[0]

    match_nis_superbias = np.where(read_patterns == 'NISRAPID')[0]
    match_nis_common = np.where(instruments == 'NIRISS')[0]
    match_nis_distortion_area = np.where((pupilnames == 'F115W') & (instruments == 'NIRISS') & (filternames == 'CLEAR'))[0]
    match_nis_flat = np.where((pupilnames == 'F115W') & (instruments == 'NIRISS') & (filternames == 'CLEAR'))[0]

    # Check that reference files that are covered by reffile_overrides
    # are equal to the override values.
    for index in match_nrc_sw_superbias:
        assert yam.info['superbias'][index] == 'my_reffiles/my_superbias_for_b4.fits'
    for index in match_nrc_lw_superbias:
        assert yam.info['superbias'][index] == 'my_reffiles/my_superbias_for_b5.fits'
    for index in match_nrc_sw_common:
        assert yam.info['linearity'][index] == 'my_reffiles/my_linearity_for_b4.fits'
        assert yam.info['saturation'][index] == 'my_reffiles/my_saturation_for_b4.fits'
        assert yam.info['gain'][index] == 'my_reffiles/my_gain_for_b4.fits'
        assert yam.info['badpixmask'][index] == 'my_reffiles/my_bpm_for_b4.fits'
    for index in match_nrc_lw_common:
        assert yam.info['linearity'][index] == 'my_reffiles/my_linearity_for_b5.fits'
        assert yam.info['saturation'][index] == 'my_reffiles/my_saturation_for_b5.fits'
        assert yam.info['gain'][index] == 'my_reffiles/my_gain_for_b5.fits'
        assert yam.info['badpixmask'][index] == 'my_reffiles/my_bpm_for_b5.fits'
    for index in match_nrc_sw_distortion_area:
        assert yam.info['astrometric'][index] == 'my_reffiles/my_distortion_for_b4.asdf'
        assert yam.info['pixelAreaMap'][index] == 'my_reffiles/my_pam_for_b4.fits'
    for index in match_nrc_lw_distortion_area:
        assert yam.info['astrometric'][index] == 'my_reffiles/my_distortion_for_b5.asdf'
        assert yam.info['pixelAreaMap'][index] == 'my_reffiles/my_pam_for_b5.fits'
    for index in match_nrc_lw_flat:
        assert yam.info['pixelflat'][index] == 'my_reffiles/my_flatfield_for_b5.fits'

    for index in match_nis_superbias:
        assert yam.info['superbias'][index] == 'my_niriss_supebias.fits'
    for index in match_nis_common:
        assert yam.info['linearity'][index] == 'my_niriss_linearity,fits'
        assert yam.info['saturation'][index] == 'my_niriss_saturation.fits'
        assert yam.info['gain'][index] == 'my_niriss_gain.fits'
        assert yam.info['badpixmask'][index] == 'my_niriss_badpixmask.fits'
    for info in match_nis_distortion_area:
        assert yam.info['astrometric'][index] == 'my_niriss_disotrtion.asdf'
        assert yam.info['pixelAreaMap'][index] == 'my_niriss_area.fits'
    for info in match_nis_flat:
        assert yam.info['pixelflat'][index] == 'my_niriss_flatfield.fits'

    # Check that reference files covered by reffile_overrides contain
    # the CRDS_PATH, which here has been set to the temp directory
    match_nrc_sw_defaults = np.where((instruments == 'NIRCAM') & (sw_filternames == 'F070W') & (detectors != 'B5'))[0]
    match_nrc_lw_defaults = np.where((instruments == 'NIRCAM') & (lw_filternames == 'F335M') & (detectors == 'B5'))[0]

    for index in list(match_nrc_sw_defaults) + list(match_nrc_lw_defaults):
        assert temp_output_dir in yam.info['superbias'][index]
        assert temp_output_dir in yam.info['astrometric'][index]
        assert temp_output_dir in yam.info['pixelAreaMap'][index]
        assert temp_output_dir in yam.info['ipc'][index]

    # Clean up
    os.system('rm -r {}'.format(temp_output_dir))


# Return environment variable to original value. This is helpful when
# calling many tests at once, some of which need the real value.
if not ON_TRAVIS:
    os.environ['MIRAGE_DATA'] = orig_mirage_data


"""Test the functions provided by mirage.psf.psf_selection

Authors
-------
    - Lauren Chambers

Use
---
    >>> pytest test_psf_selection.py
"""
import os
import shutil

from astropy.io import fits
import photutils
import pytest

from mirage.psf.psf_selection import get_library_file, _load_itm_library
from mirage.utils.utils import ensure_dir_exists

# Define directory and file locations
__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
TEMP_TEST_DIRECTORY = os.path.join(__location__, 'temp_data', 'test_psf_selection')


@pytest.fixture(scope="module")
def test_directory(test_dir=TEMP_TEST_DIRECTORY):
    """Create a test directory.

    Parameters
    ----------
    test_dir : str
        Path to directory used for testing

    Yields
    -------
    test_dir : str
        Path to directory used for testing
    """
    # Create directory and yield its name
    ensure_dir_exists(test_dir)  # creates directory with default mode=511
    yield test_dir

    # Remove directory
    if os.path.isdir(test_dir):
        shutil.rmtree(test_dir)


@pytest.fixture(scope="module")
def test_segment_psf_library_file(test_directory):
    """Download and save locally a segment PSF library file from Box for testing

    Yields
    -------
    test_lib_filename : str
        Path to test segment PSF library used for testing
    """
    # Download file and yield its name
    test_lib_filename = os.path.join(test_directory,
                                     'nircam_nrca3_f212n_fovp1024_samp1_npsf1_seg12.fits')
    url = "https://stsci.box.com/shared/static/4c0em1yhb1qsvrpku7j0jw1tztucvkih.fits"
    with fits.open(url) as hdulist:
        hdulist.writeto(test_lib_filename)
        yield test_lib_filename


@pytest.fixture(scope="module")
def itm_file(test_directory):
    """Download and save locally an ITM image file from Box for testing

    Yields
    -------
    itm_filename : str
        Path to ITM image file used for testing
    """
    # Download file and yield its name
    itm_filename = os.path.join(test_directory,
                                     'itm_file.fits')
    url = "https://stsci.box.com/shared/static/lsvb6acrlubwnlzz7wkkpbjgm01bbzgb.fits"
    with fits.open(url) as hdulist:
        hdulist.writeto(itm_filename)
        yield itm_filename


def test_get_segment_psf_library_file(test_segment_psf_library_file):
    """Test the identification of a segment PSF library file

    Parameters
    ----------
    test_segment_psf_library_file : str
        Path to test segment PSF library used for testing
    """
    instrument = "NIRCam"
    detector = "NRCA3"
    filt = "F212N"
    pupil = "CLEAR"
    wfe = ''
    wfe_group = 0
    library_path = os.path.dirname(test_segment_psf_library_file)
    segment_id = 12

    # Ensure the right file is found given the right parameters
    match_file = get_library_file(instrument, detector, filt, pupil, wfe, wfe_group,
                     library_path, wings=False, segment_id=segment_id)
    assert match_file == test_segment_psf_library_file, \
        "Failed to match segment PSF library"

    # Ensure an error is raised if no file exists that matches
    with pytest.raises(ValueError) as e:
        detector = "NRCB5"
        segment_id = 11
        get_library_file(instrument, detector, filt, pupil, wfe, wfe_group,
                         library_path, wings=False, segment_id=segment_id)
        assert "No PSF library file found matching requested parameters" in e


def test_load_itm_library(itm_file):
    """Test the loading of a GriddedPSFModel object from an ITM FITS file

    Parameters
    ----------
    itm_filename : str
        Path to ITM image file used for testing
    """
    lib_model = _load_itm_library(itm_file)

    assert isinstance(lib_model, photutils.psf.models.GriddedPSFModel), \
        'ITM PSF library not created correctly'
    assert lib_model.grid_xypos == [(1023.0, 1023.0)], \
        'ITM PSF library not created correctly'
    assert lib_model.oversampling == 1, \
        'ITM PSF library not created correctly'
    assert lib_model.data.shape == (1, 2048, 2048), \
        'ITM PSF library not created correctly'



#! /usr/bin/env python

"""This module contains code for reading and writing source spectra into
hdf5 files. This format is needed as input to the disperser package
used to create dispersed seed images

Authors
-------

    - Bryan Hilbert

Use
---

    This script is intended to be executed as such:

    ::

        from mirage.catalogs import hdf5_catalog
        spectra_dict = hdf5_catalog.open('my_catalog_file.hdf5')

        hdf5_catalog.save(spectra_dict, 'my_updated_catalog_file.hdf5')
"""

import astropy.units as u
import h5py

from mirage.utils.constants import FLAMBDA_CGS_UNITS, FLAMBDA_MKS_UNITS, FNU_CGS_UNITS, FNU_MKS_UNITS


def open(filename):
    """Read in contents of an hdf5 file

    Parameters
    ----------
    filename : str
        Name of file to be opened

    Returns
    -------
    contents : dict
        Dictionary containing the contents of the file
        Dictionary format:
        keys are the index numbers of the sources corresponding to the segmentation map
        Each value is a dictionary containing keys 'wavelengths' and 'fluxes'.
        'wavelengths' is an astropy.units Quantity composed of a list of wavelength values
        and a wavelength unit
        'fluxes' is an astropy.units Quantity composed of a list of flux values with flux unit
    """
    contents = {}
    with h5py.File(filename, 'r') as file_obj:
        no_wave_units = False
        no_flux_units = False
        for key in file_obj.keys():
            dataset = file_obj[key]
            try:
                wave_units_string = dataset.attrs['wavelength_units']
            except KeyError:
                wave_units_string = 'micron'
                no_wave_units = True
            try:
                flux_units_string = dataset.attrs['flux_units']
            except KeyError:
                flux_units_string = 'flam_cgs'
                no_flux_units = True

            # Catch common errors
            if wave_units_string.lower() in ['microns', 'angstroms', 'nanometers']:
                wave_units_string = wave_units_string[0:-1]

            # Convert the unit strings into astropy.units Unit object
            wave_units = string_to_units(wave_units_string)
            flux_units = string_to_units(flux_units_string)

            # Get the data
            waves = dataset[0] * wave_units
            fluxes = dataset[1] * flux_units

            # Convert wavelengths to microns and flux values to f_lambda in cgs
            if wave_units != u.micron:
                if wave_units.is_equivalent(u.micron):
                    waves = waves.to(u.micron)
                elif wave_units.is_equivalent(u.Hz):
                    waves = waves.to(u.micron, equivalencies=u.spectral())
                else:
                    raise ValueError("Wavelength units of {} in dataset {} are not compatible with microns."
                                     .format(wave_units, key))
            if flux_units != FLAMBDA_CGS_UNITS:
                if flux_units.is_equivalent(FLAMBDA_CGS_UNITS):
                    fluxes = fluxes.to(FLAMBDA_CGS_UNITS)
                elif flux_units.is_equivalent(FNU_CGS_UNITS):
                    fluxes = fluxes.to(FLAMBDA_CGS_UNITS, u.spectral_density(waves))
                elif flux_units == u.pct:
                    pass
                else:
                    raise ValueError("Flux density units of {} in dataset {} are not compatible with f_lambda."
                                     .format(flux_units, key))

            contents[int(key)] = {'wavelengths': waves, 'fluxes': fluxes}
    if no_wave_units:
        print("{}: No wavelength units provided. Assuming MIRCONS.".format(filename))
    if no_flux_units:
        print("{}: No flux density units provided. Assuming F_lambda in CGS (erg/sec/cm^2/A)".format(filename))
    return contents


def open_tso(filename):
    """Read in an hdf5 file containing TSO data

    Parameters
    ----------
    filename : str
        Name of file to be opened

    Returns
    -------
    contents : dict
        Dictionary containing the contents of the file
        Dictionary format:
        keys are the index numbers of the sources corresponding to the segmentation map
        Each value is a dictionary containing keys 'times' and 'fluxes'.
        'times' is an astropy.units Quantity composed of a list of time values
        and a time unit
        'fluxes' is an astropy.units Quantity composed of a list of flux values with flux unit
    """
    contents = {}
    with h5py.File(filename, 'r') as file_obj:
        no_time_units = False
        no_flux_units = False
        for key in file_obj.keys():
            dataset = file_obj[key]
            try:
                time_units_string = dataset.attrs['time_units']
            except KeyError:
                time_units_string = 'second'
                no_time_units = True
            try:
                flux_units_string = dataset.attrs['flux_units']
            except KeyError:
                flux_units_string = 'normalized'
                no_flux_units = True

            # Catch common errors
            if time_units_string.lower() in ['seconds', 'minutes', 'hours', 'days']:
                time_units_string = time_units_string[0:-1]

            # Convert the unit strings into astropy.units Unit object
            time_units = string_to_units(time_units_string)
            flux_units = string_to_units(flux_units_string)

            # Get the data
            times = dataset[0] * time_units
            fluxes = dataset[1] * flux_units

            # Convert times to seconds
            if time_units != u.second:
                if time_units.is_equivalent(u.second):
                    times = times.to(u.second)
                else:
                    raise ValueError("Time units of {} in dataset {} are not compatible with seconds."
                                     .format(time_units, key))
            if flux_units != u.pct:
                raise ValueError("Flux units are assumed to be normalized. But units in dataset {} are {}."
                                 .format(key, flux_units))

            contents[int(key)] = {'times': times, 'fluxes': fluxes}
    if no_time_units:
        print("{}: No time units provided. Assuming SECONDS.".format(filename))
    if no_flux_units:
        print("{}: No flux density units provided. Assuming light curve is normalized.".format(filename))
    return contents


def save(contents, filename, wavelength_unit='', flux_unit=''):
    """Save a dictionary into an hdf5 file

    Paramters
    ---------
    contents : dict
        Dictionary of data. Dictionary format:
        keys are the index numbers of the sources corresponding to the segmentation map
        Each value is a dictionary containing keys 'wavelengths' and 'fluxes'.
        'wavelengths' is an astropy.units Quantity composed of a list of wavelength values
        and wavelength units
        'fluxes' is an astropy.units Quantity composed of a list of flux values with flux units

    filename : str
        Name of hdf5 file to produce
    """
    with h5py.File(filename, "w") as file_obj:
        for key in contents.keys():
            flux = contents[key]['fluxes']
            wavelength = contents[key]['wavelengths']

            # If units are astropy.units Units objects, change to strings
            # If wavelengths are not a Quantity, fall back onto wavelength_unit
            # and flux_unit
            if isinstance(wavelength, u.quantity.Quantity):
                wavelength_units = units_to_string(wavelength.unit)
                wavelength_values = wavelength.value
            else:
                wavelength_units = wavelength_unit
                wavelength_values = wavelength
            if isinstance(flux, u.quantity.Quantity):
                flux_units = units_to_string(flux.unit)
                flux_values = flux.value
            else:
                flux_units = flux_unit
                flux_values = flux

            dset = file_obj.create_dataset(str(key), data=[wavelength_values, flux_values], dtype='f',
                                           compression="gzip", compression_opts=9)

            # Set dataset units. Not currently inspected by mirage.
            if wavelength_units != '':
                dset.attrs[u'wavelength_units'] = wavelength_units
            if flux_units != '':
                dset.attrs[u'flux_units'] = flux_units


def save_tso(contents, filename, time_unit='', flux_unit=''):
    """Save a dictionary of TSO data into an hdf5 file

    Paramters
    ---------
    contents : dict
        Dictionary of data. Dictionary format:
        keys are the index numbers of the sources corresponding to the segmentation map
        Each value is a dictionary containing keys 'times' and 'fluxes'.
        'times' is an astropy.units Quantity composed of a list oftime values
        and time units
        'fluxes' is an astropy.units Quantity composed of a list of flux values with flux units

    filename : str
        Name of hdf5 file to produce

    time_unit : str
        In the case where the ``times`` entry of ``contents`` is a numpy array
        rather than an astropy Quantity, this unit will be used for the ``times``
        entry. (e.g. 'second')

    flux_unit : str
        In the case where the ``fluxes`` entry of ``contents`` is a numpy array
        rather than an astropy Quantity, this unit will be used for the ``fluxes``
        entry. (e.g. 'flam_cgs')
    """
    with h5py.File(filename, "w") as file_obj:
        for key in contents.keys():
            flux = contents[key]['fluxes']
            time = contents[key]['times']

            # If units are astropy.units Units objects, change to strings
            # If wavelengths are not a Quantity, fall back onto wavelength_unit
            # and flux_unit
            if isinstance(time, u.quantity.Quantity):
                time_units = units_to_string(time.unit)
                time_values = time.value
            else:
                time_units = time_unit
                time_values = time
            if isinstance(flux, u.quantity.Quantity):
                flux_units = units_to_string(flux.unit)
                flux_values = flux.value
            else:
                flux_units = flux_unit
                flux_values = flux

            dset = file_obj.create_dataset(str(key), data=[time_values, flux_values], dtype='f',
                                           compression="gzip", compression_opts=9)

            # Set dataset units. Not currently inspected by mirage.
            if time_units != '':
                dset.attrs[u'time_units'] = time_units
            if flux_units != '':
                dset.attrs[u'flux_units'] = flux_units


def string_to_units(unit_string):
    """Convert a string containing units to an astropy.units Quantity

    Parameters
    ----------
    unit_string : str
        String containing units (e.g. 'erg/sec/cm/cm/A')

    Returns
    -------
    units : astropy.units Quantity
    """
    if unit_string.lower() in ['flam', "flam_cgs"]:
        return FLAMBDA_CGS_UNITS
    elif unit_string.lower() == 'flam_mks':
        return FLAMBDA_MKS_UNITS
    elif unit_string.lower() in ['fnu', 'fnu_cgs']:
        return FNU_CGS_UNITS
    elif unit_string.lower() == 'fnu_mks':
        return FNU_MKS_UNITS
    elif unit_string in ['normalized', 'NORMALIZED', 'Normalized']:
        return u.pct
    else:
        try:
            return u.Unit(unit_string)
        except ValueError as e:
            print(e)


def units_to_string(unit):
    """Convert the units of an astropy.units Quantity to a string

    Parameters
    ----------
    quantity : astropy.units Quantity

    Returns
    -------
    unit_string : str
        String representation of the units in quantity
    """
    if unit == FLAMBDA_CGS_UNITS:
        return 'flam_cgs'
    elif unit == FLAMBDA_MKS_UNITS:
        return 'flam_mks'
    elif unit == FNU_CGS_UNITS:
        return 'fnu_cgs'
    elif unit == FNU_MKS_UNITS:
        return 'fnu_mks'
    elif unit == u.pct:
        return 'normalized'
    else:
        return unit.to_string()



"""System test of mirage/FGS that creates simulations based on a .yaml file.

Authors
-------
    - Johannes Sahlmann

Use
---
    >>> pytest -s test_fgs_imaging.py


"""

import os
import pytest

from mirage import imaging_simulator as im

os.environ['TEST_FGS_DATA'] = os.path.join(os.path.dirname(__file__), 'test_data/FGS')

# Determine if tests are being run on Travis
ON_TRAVIS = 'travis' in os.path.expanduser('~')


@pytest.mark.skipif(ON_TRAVIS,
                    reason="Cannot access mirage data in the central storage directory from Travis CI.")
def test_fgs_imaging():
    m = im.ImgSim(offline=True)
    m.paramfile = os.path.join(os.path.dirname(__file__), 'test_data/FGS/fgs_imaging_example.yaml')
    m.create()



#! /usr/bin/env python

"""Utility for determining how to split up seed images and dark files
that are too large. This is meant to roughly mimic the way that the
DMS pipeline will do the same job. Files that are spllit will have
``segNNN`` added to their names, where ``NNN`` is a number.
"""
import copy
import numpy as np
from mirage.utils.constants import FILE_SPLITTING_LIMIT


def find_file_splits(xdim, ydim, groups, integrations, frames_per_group=None,
                     pixel_limit=FILE_SPLITTING_LIMIT):
    """Determine the frame and/or integration numbers where a file
    should be split in order to keep the file size reasonable.

    Parameters
    ----------
    xdim : int
        Number of columns in the aperture

    ydim : int
        Number of rows in the aperture

    groups : int
        Number of groups (or frames) in an integration

    integrations : int
        Number of integrations in exposure

    frames_per_group : int
        Number of frames (inculding skipped frames) per
        group

    pixel_limit : int
        Proxy for file size limit. Number of pixel read outs to treat
        as the upper limit to be contained in a single file.

    Returns
    -------
    split : bool
        Whether or not the exposure needs to be split into multiple
        files

    group_list : numpy.ndarray
        1D array listing the beginning group number of each file split

    integration_list : numpy.ndarray
        1d array listing the beginning integration number of each file
        split
    """
    pix_per_group = ydim * xdim
    pix_per_int = groups * pix_per_group
    observation = pix_per_int * integrations

    # Default = no splitting
    split = False
    group_list = np.array([0, groups])
    integration_list = np.array([0, integrations])

    # Check for splitting between groups first
    # i.e. splitting within an integration
    if pix_per_int > pixel_limit:
        split = True
        delta_group = np.int(pixel_limit / pix_per_group)
        # If calculations are being done using frames rather than groups,
        # make sure that we don't split in the middle of a group. Force
        # the splits to be between groups.
        if frames_per_group is not None:
            if delta_group % frames_per_group != 0:
                closest_multiple = delta_group // frames_per_group
                delta_group = closest_multiple * frames_per_group

        group_list = np.arange(0, groups, delta_group).astype(np.int)
        group_list = np.append(group_list, groups)
        print('Splitting within each integration:')
        integration_list = np.arange(integrations + 1).astype(np.int)
        print('integration_list: ', integration_list)
        print('group_list: ', group_list)
    elif observation > pixel_limit:
        split = True
        print('Splitting by integration:')
        group_list = np.array([0, groups])
        delta_int = np.int(pixel_limit / pix_per_int)
        integration_list = np.arange(0, integrations, delta_int).astype(np.int)
        integration_list = np.append(integration_list, integrations)
        print('integration_list: ', integration_list)
        print('group_list: ', group_list)

    return split, group_list, integration_list


class SplitFileMetaData:
    def __init__(self, integration_splits, frame_splits, DMS_integration_splits, DMS_group_splits,
                 frames_per_integration, frames_per_group, frametime):
        """Calculate metadata values related to the position of the split files'
        data within the segment and exposure. These data will be needed when
        reconstructing `parts` into segments later

        Parameters
        ----------
        integration_splits : list
            List of the beginning integration numbers for file splitting
            when split using Mirage's parameters (i.e. including 'parts')
            This list is generated using ``find_file_splits``

        frame_splits : list
            List of the beginning frame numbers for file splitting when
            split using Mirage's parameters (i.e. including 'parts')
            This list is generated using ``find_file_splits``

        DMS_integration_splits : list
            List of the beginning integration numbers for file splitting
            when split using DMS parameters (i.e. no splits within an
            integration, splitting on groups rather than frames)
            This list is generated using ``find_file_splits``

        DMS_group_splits : list
            List of the beginning group numbers for file splitting
            when split using DMS parameters (i.e. no splits within an
            integration, splitting on groups rather than frames)
            This should always be [0, last_group]
            This list is generated using ``find_file_splits``

        frames_per_integration : int
            Number of frames per integration of the simulated data

        frames_per_group : int
            Number of frames per group in the simulated data

        frametime : float
            Exposure time of one frame (e.e. detector readout) of the
            simulated data
        """
        self.total_frames = []
        self.total_ints = []
        self.time_start = []
        self.frame_start = []
        self.segment_part_number = []
        self.segment_ints = []
        self.segment_frames = []
        self.segment_frame_start_number = []
        self.segment_int_start_number = []
        self.part_int_start_number = []
        self.part_frame_start_number = []
        self.segment_number = []
        segment_starting_int_number = 0
        segment_part_number = 0
        segment_frame_start_number = 0
        segment_int_start_number = 0

        i = 1
        previous_segment = 0
        for int_start in integration_splits[:-1]:
            int_end = integration_splits[i]

            j = 1
            for initial_frame in frame_splits[:-1]:
                frame_start = int_start * (frames_per_integration + 1) + initial_frame
                frame_end = frame_start + (frame_splits[j] - initial_frame)
                time_start = frame_start * frametime
                time_end = frame_end * frametime
                total_ints = int_end - int_start
                # total_frames should be the total number of frames
                # in the chunk to be worked on, including resets
                total_frames = (frame_end - frame_start) * total_ints
                if total_ints > 1:
                    total_frames += total_ints - 1

                #print('int_start, int_end, i:', int_start, int_end, i)
                #print('initial_frame, frame_start, frame_end, total_frames:', initial_frame, frame_start, frame_end, total_frames)

                self.total_frames.append(total_frames)
                self.total_ints.append(total_ints)
                self.time_start.append(time_start)
                self.frame_start.append(frame_start)

                segment_number = np.where(int_end <= DMS_integration_splits)[0][0]
                self.segment_ints.append(DMS_integration_splits[segment_number] - DMS_integration_splits[segment_number - 1])
                self.segment_frames.append((DMS_group_splits[1] - DMS_group_splits[0]) * frames_per_group)

                if segment_number == previous_segment:
                    segment_part_number += 1
                    part_int_start_number = int_start - segment_starting_int_number
                    part_frame_start_number = initial_frame
                else:
                    segment_part_number = 1
                    previous_segment = copy.deepcopy(segment_number)
                    segment_frame_start_number = initial_frame
                    segment_int_start_number = int_start
                    part_int_start_number = 0
                    part_frame_start_number = 0
                    segment_starting_int_number = copy.deepcopy(int_start)

                self.segment_part_number.append(segment_part_number)
                self.segment_frame_start_number.append(segment_frame_start_number)
                self.segment_int_start_number.append(segment_int_start_number)
                self.part_int_start_number.append(part_int_start_number)
                self.part_frame_start_number.append(part_frame_start_number)
                self.segment_number.append(segment_number)

                #print("\n\nSegment Numbers:")
                #print(int_start, initial_frame)
                #print(self.total_frames)
                #print(self.segment_number, self.segment_part_number)
                #print(self.segment_ints, self.segment_frames)
                #print(self.segment_int_start_number, self.segment_frame_start_number)
                #print(self.part_int_start_number, self.part_frame_start_number)
                #print('\n\n')

                j += 1

            i += 1



#! /usr/bin/env python

'''
Segmentation map creation. Developed in conjunction with
the seed image generator code for catalogs
catalog_seed_image.py
'''

import numpy as np


class SegMap():
    def __init__(self):
        self.xdim = 2048
        self.ydim = 2048
        self.zdim = None

    def initialize_map(self):
        if self.zdim is None:
            self.segmap = np.zeros((self.ydim, self.xdim), dtype=np.int64)
        else:
            self.segmap = np.zeros((self.zdim, self.ydim, self.xdim), dtype=np.int64)

    def add_object_basic(self, ystart, yend, xstart, xend, number):
        # Add an object to the segmentation map
        # in simplest way possible. All pixels
        # in the box are set to the index number
        # regardless of signal
        self.segmap[ystart:yend, xstart:xend] = number

    def add_object_perccut(self, image, ystart, xstart, number, perc):
        # Add an object to the segmentation map
        # In this case, only flag pixels whose
        # signal is higher than the given cutoff
        yd, xd = image.shape
        stamp = self.segmap[ystart:ystart+yd, xstart:xstart+xd]
        maxsig = np.max(image)
        cutoff = maxsig * perc
        flag = image >= cutoff
        stamp[flag] = number

    def add_object_noise(self, image, ystart, xstart, number, noise):
        # Add an object to the segmentation map
        # In this case, only flag pixels whose
        # signal is higher than a calculated cutoff
        # that is based on average noise/background
        # values for NIRCam
        yd, xd = image.shape
        stamp = self.segmap[ystart:ystart+yd, xstart:xstart+xd]
        flag = image >= noise
        stamp[flag] = number



#! /usr/bin/env python

'''
To make the generation of imaging (including moving target)
simulated integrations easier, combine the 3 relevant stages
of the simulator (seed image generator, dark prep,
obervation generator) into a single script.

Inputs:
paramfile - Name of yaml file to be used as simulator input.
            For details  on the information contained in the
            yaml files, see the readme file associated with
            the mirage github repo:
            https://github.com/spacetelescope/mirage.git

override_dark - If you wish to use a dark current file that
                has already gone through the dark_prep step
                of the pipeline and wish to use that for the
                simulation, set override_dark equal to the
                dark's filename. The dark_prep step will then
                be skipped.

HISTORY:
13 November 2017 - created, Bryan Hilbert
13 July 2018 - updated for name change to Mirage, Bryan Hilbert
'''

import os
import argparse

from .seed_image import catalog_seed_image
from .dark import dark_prep
from .ramp_generator import obs_generator
from .utils import read_fits
from .utils.utils import expand_environment_variable


class ImgSim():
    def __init__(self, paramfile=None, override_dark=None, offline=False):
        self.env_var = 'MIRAGE_DATA'
        datadir = expand_environment_variable(self.env_var, offline=offline)

        self.paramfile = paramfile
        self.override_dark = override_dark
        self.offline = offline

    def create(self):
        # Create seed image
        cat = catalog_seed_image.Catalog_seed(offline=self.offline)
        cat.paramfile = self.paramfile
        cat.make_seed()

        # Create observation generator object
        obs = obs_generator.Observation(offline=self.offline)

        # Prepare dark current exposure if
        # needed.
        if self.override_dark is None:
            print('Perform dark preparation:')
            d = dark_prep.DarkPrep(offline=self.offline)
            d.paramfile = self.paramfile
            d.prepare()

            if len(d.dark_files) == 1:
                obs.linDark = d.prepDark
            else:
                obs.linDark = d.dark_files
        else:
            self.read_dark_product(self.override_dark)
            obs.linDark = self.prepDark

        # Combine into final observation
        obs.paramfile = self.paramfile
        if len(cat.seed_files) == 1:
            obs.seed = cat.seedimage
            obs.segmap = cat.seed_segmap
            obs.seedheader = cat.seedinfo
        else:
            obs.seed = cat.seed_files
            print('USING SEED_FILES')
            print(cat.seed_files)
        obs.create()

        # Make useful information class attributes
        self.seedimage = cat.seedimage
        self.seed_segmap = cat.seed_segmap
        self.seedinfo = cat.seedinfo
        self.linDark = obs.linDark

    def read_dark_product(self, file):
        # Read in dark product that was produced
        # by dark_prep.py
        self.prepDark = read_fits.Read_fits()
        self.prepDark.file = file
        self.prepDark.read_astropy()

    def add_options(self, parser=None, usage=None):
        if parser is None:
            parser = argparse.ArgumentParser(usage=usage, description="Wrapper for the creation of WFSS simulated exposures.")
        parser.add_argument("paramfile", help='Name of simulator input yaml file')
        parser.add_argument("--override_dark", help="If supplied, skip the dark preparation step and use the supplied dark to make the exposure", default=None)
        return parser


if __name__ == '__main__':

    usagestring = 'USAGE: imaging_simualtor.py file1.yaml'

    obs = ImgSim()
    parser = obs.add_options(usage=usagestring)
    args = parser.parse_args(namespace=obs)
    obs.create()



'''A collection of basic routines for doing rotation calculations
Changed to express quaternion as vector before scalar part(a, p, )
'''

from math import radians, cos, sin, degrees, asin, atan2, sqrt, acos

import numpy as np


def unit(ra, dec):
    ''' Converts vector expressed in Euler angles to unit vector components.
    ra and dec in degrees
    Can be used for V2V3 after converting from arcsec to degrees)'''

    rar = radians(ra)
    decr = radians(dec)
    u = np.array([cos(rar)*cos(decr), sin(rar)*cos(decr), sin(decr)])
    return u


def radec(u):
    '''convert unit vector to Euler angles
    u is an array or list of length 3'''

    if len(u) != 3:
        print('Not a vector')
        return
    norm = sqrt(u[0]**2 + u[1]**2 + u[2]**2) # Works for list or array
    dec = degrees(asin(u[2]/norm))
    ra = degrees(atan2(u[1], u[0])) # atan2 puts it in the correct quadrant
    if ra < 0.0: ra += 360.0    # Astronomers prefer the range 0 to 360 degrees
    return (ra, dec)


def v2v3(u):
    '''Convert unit vector to v2v3'''
    if len(u) != 3:
        print ('Not a vector')
        return
    norm = sqrt(u[0]**2 + u[1]**2 + u[2]**2) # Works for list or array
    v2 = 3600*degrees(atan2(u[1], u[0])) # atan2 puts it in the correct quadrant
    v3 = 3600*degrees(asin(u[2]/norm))
    return (v2, v3)


def rotate(axis, angle):
    '''Fundamental rotation matrices.
    Rotate by angle measured in degrees, about axis 1 2 or 3'''
    if axis not in list(range(1, 4)):
        print ('Axis must be in range 1 to 3')
        return
    r = np.zeros((3, 3))
    ax0 = axis-1 #Allow for zero offset numbering
    theta = radians(angle)
    r[ax0, ax0] = 1.0
    ax1 = (ax0+1) % 3
    ax2 = (ax0+2) % 3
    r[ax1, ax1] = cos(theta)
    r[ax2, ax2] = cos(theta)
    r[ax1, ax2] = -sin(theta)
    r[ax2, ax1] = sin(theta)
    return r


def rv(v2, v3):
    '''Rotate from v2, v3 position to V1 axis'''
    v2d = v2/3600.0 #convert from arcsec to degrees
    v3d = v3/3600.0
    mv2 = rotate(3, -v2d)
    mv3 = rotate(2, v3d)
    rv = np.dot(mv3, mv2)
    return rv


def slew(v2t, v3t, v2a, v3a):
    ''' Calculate matrix which slews from target (v2t, v3t)
    to aperture position (v2a, v3a)'''
    v2td = v2t/3600.0
    v3td = v3t/3600.0
    v2ad = v2a/3600.0
    v3ad = v3a/3600.0
    r1 = rotate(3, -v2td)
    r2 = rotate(2, v3td-v3ad)
    r3 = rotate(3, v2ad)
    # Combine r3 r2 r1
    mv = np.dot(r2, r1)
    mv = np.dot(r3, mv)
    return mv


def attitude(v2, v3, ra, dec, pa):
    '''This will make a rotation matrix which rotates a unit vector representing a v2, v3 position
    to a unit vector representing an RA, Dec pointing with an assigned position angle
    Described in JWST-STScI-001550, SM-12, section 6.1'''

    # v2, v3 in arcsec, ra, dec and position angle in degrees
    v2d = v2/3600.0
    v3d = v3/3600.0

    # Get separate rotation matrices
    mv2 = rotate(3, -v2d)
    mv3 = rotate(2, v3d)
    mra = rotate(3, ra)
    mdec = rotate(2, -dec)
    mpa = rotate(1, -pa)

    # Combine as mra*mdec*mpa*mv3*mv2
    m = np.dot(mv3, mv2)
    m = np.dot(mpa, m)
    m = np.dot(mdec, m)
    m = np.dot(mra, m)

    return m


def pointing(attitude, v2, v3):
    '''Using the attitude matrix to calculate where any v2v3 position points on the sky'''
    v2d = v2/3600.0
    v3d = v3/3600.0
    v = unit(v2d, v3d)
    w = np.dot(attitude, v)
    rd = radec(w)
    return rd # tuple containing ra and dec


def getv2v3(attitude, ra, dec):
    ''' Using the inverse of attitude matrix
    find v2, v3 position of any RA and DEC'''
    urd = unit(ra, dec)
    invAtt = np.linalg.inv(attitude)
    uv = np.dot(invAtt, urd)
    (v2, v3) = radec(uv)
    if v2 > 180.0: v2 = v2-360.0
    v2 = 3600.0*v2
    v3 = 3600.0*v3
    return (v2, v3)


def posangle(attitude, v2, v3):
    ''' Using the attitude matrix find the V3 angle at arbitrary v2, v3
    This is the angle measured from North to V3 in an anti-clockwise direction
    i.e. North to East
    Formulae from JWST-STScI-001550, SM-12, section 6.2
    Subtract 1 from each index in the text to allow for python zero indexing'''

    A = attitude  # Synonym to simplify typing
    v2r = radians(v2/3600.0)
    v3r = radians(v3/3600.0)
    x = -(A[2, 0]*cos(v2r) + A[2, 1]*sin(v2r))*sin(v3r) + A[2, 2]*cos(v3r)
    y = (A[0, 0]*A[1, 2] - A[1, 0]*A[0, 2])*cos(v2r) + (A[0, 1]*A[1, 2] - A[1, 1]*A[0, 2])*sin(v2r)
    pa = degrees(atan2(y, x))
    return pa


def rodrigues(attitude):
    '''Interpret rotation matrix as a single rotation by angle phi around unit length axis
    Return axis, angle and matching quaternion'''

    A = attitude # Synonym for clarity and to save typing
    cos_phi = 0.5*(A[0, 0] + A[1, 1] + A[2, 2] - 1.0)
    phi = acos(cos_phi)
    axis = np.array([A[2, 1]-A[1, 2], A[0, 2]-A[2, 0], A[1, 0]-A[0, 1]])/(2.0*sin(phi))
    # Make corresponding quaternion
    q = np.hstack(( axis*sin(phi/2.0), [cos(phi/2.0)]))
    phi = degrees(phi)
    return (axis, phi, q)


def cross(a, b):
    """cross product of two vectors"""
    c = np.array([a[1]*b[2]-a[2]*b[1], a[2]*b[0]-a[0]*b[2], a[0]*b[1]-a[1]*b[0]])
    return c


def axial(ax, phi, u):
    ''' Apply direct rotation to a vector using Rodrigues' formula
    ax is unit axis vector  phi is rotation angle in degrees
    u is initial vector'''
    rphi = radians(phi)
    v = u*cos(rphi) + cross(ax, u)*sin(rphi) + ax*np.dot(ax, u)*(1-cos(rphi))
    return v
